# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Python camera library for the Rasperry-Pi camera module
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    print_function,
    division,
    absolute_import,
    )

# Make Py2's str and range equivalent to Py3's
str = type('')

import datetime
import threading
import warnings
import ctypes as ct

from . import bcm_host, mmal, mmalobj as mo
from .frames import PiVideoFrame, PiVideoFrameType
from .exc import (
    PiCameraMMALError,
    PiCameraValueError,
    PiCameraIOError,
    PiCameraRuntimeError,
    PiCameraResizerEncoding,
    PiCameraAlphaStripping,
    PiCameraResolutionRounded,
    )


class PiEncoder(object):
    """
    Base implementation of an MMAL encoder for use by PiCamera.

    The *parent* parameter specifies the :class:`PiCamera` instance that has
    constructed the encoder. The *camera_port* parameter provides the MMAL
    camera port that the encoder should enable for capture (this will be the
    still or video port of the camera component). The *input_port* parameter
    specifies the MMAL port that the encoder should connect to its input.
    Sometimes this will be the same as the camera port, but if other components
    are present in the pipeline (e.g. a splitter), it may be different.

    The *format* parameter specifies the format that the encoder should
    produce in its output. This is specified as a string and will be one of
    the following for image encoders:

    * ``'jpeg'``
    * ``'png'``
    * ``'gif'``
    * ``'bmp'``
    * ``'yuv'``
    * ``'rgb'``
    * ``'rgba'``
    * ``'bgr'``
    * ``'bgra'``

    And one of the following for video encoders:

    * ``'h264'``
    * ``'mjpeg'``

    The *resize* parameter is either ``None`` (indicating no resizing
    should take place), or a ``(width, height)`` tuple specifying the
    resolution that the output of the encoder should be resized to.

    Finally, the *options* parameter specifies additional keyword arguments
    that can be used to configure the encoder (e.g. bitrate for videos, or
    quality for images).

    .. attribute:: camera_port

        The :class:`~mmalobj.MMALVideoPort` that needs to be activated and
        deactivated in order to start/stop capture. This is not necessarily the
        port that the encoder component's input port is connected to (for
        example, in the case of video-port based captures, this will be the
        camera video port behind the splitter).

    .. attribute:: encoder

        The :class:`~mmalobj.MMALComponent` representing the encoder, or
        ``None`` if no encoder component has been created (some encoder classes
        don't use an actual encoder component, for example
        :class:`PiRawImageMixin`).

    .. attribute:: event

        A :class:`threading.Event` instance used to synchronize operations
        (like start, stop, and split) between the control thread and the
        callback thread.

    .. attribute:: exception

        If an exception occurs during the encoder callback, this attribute is
        used to store the exception until it can be re-raised in the control
        thread.

    .. attribute:: format

        The image or video format that the encoder is expected to produce. This
        is equal to the value of the *format* parameter.

    .. attribute:: input_port

        The :class:`~mmalobj.MMALVideoPort` that the encoder should be
        connected to.

    .. attribute:: output_port

        The :class:`~mmalobj.MMALVideoPort` that produces the encoder's output.
        In the case no encoder component is created, this should be the
        camera/component output port responsible for producing data. In other
        words, this attribute **must** be set on initialization.

    .. attribute:: outputs

        A mapping of ``key`` to ``(output, opened)`` tuples where ``output``
        is a file-like object, and ``opened`` is a bool indicating whether or
        not we opened the output object (and thus whether we are responsible
        for eventually closing it).

    .. attribute:: outputs_lock

        A :func:`threading.Lock` instance used to protect access to
        :attr:`outputs`.

    .. attribute:: parent

        The :class:`PiCamera` instance that created this PiEncoder instance.

    .. attribute:: pool

        A pointer to a pool of MMAL buffers.

    .. attribute:: resizer

        The :class:`~mmalobj.MMALResizer` component, or ``None`` if no resizer
        component has been created.
    """

    DEBUG = 0
    encoder_type = None

    def __init__(
            self, parent, camera_port, input_port, format, resize, **options):
        self.parent = parent
        self.encoder = None
        self.resizer = None
        self.camera_port = camera_port
        self.input_port = input_port
        self.output_port = None
        self.outputs_lock = threading.Lock() # protects access to self.outputs
        self.outputs = {}
        self.exception = None
        self.event = threading.Event()
        try:
            if parent and parent.closed:
                raise PiCameraRuntimeError("Camera is closed")
            if resize:
                self._create_resizer(*mo.to_resolution(resize))
            self._create_encoder(format, **options)
            if self.encoder:
                self.encoder.connection.enable()
            if self.resizer:
                self.resizer.connection.enable()
        except:
            self.close()
            raise

    def _create_resizer(self, width, height):
        """
        Creates and configures an :class:`~mmalobj.MMALResizer` component.

        This is called when the initializer's *resize* parameter is something
        other than ``None``. The *width* and *height* parameters are passed to
        the constructed resizer. Note that this method only constructs the
        resizer - it does not connect it to the encoder. The method sets the
        :attr:`resizer` attribute to the constructed resizer component.
        """
        self.resizer = mo.MMALResizer()
        self.resizer.inputs[0].connect(self.input_port)
        self.resizer.outputs[0].copy_from(self.resizer.inputs[0])
        self.resizer.outputs[0].format = mmal.MMAL_ENCODING_I420
        self.resizer.outputs[0].framesize = (width, height)
        self.resizer.outputs[0].commit()

    def _create_encoder(self, format):
        """
        Creates and configures the :class:`~mmalobj.MMALEncoder` component.

        This method only constructs the encoder; it does not connect it to the
        input port. The method sets the :attr:`encoder` attribute to the
        constructed encoder component, and the :attr:`output_port` attribute to
        the encoder's output port (or the previously constructed resizer's
        output port if one has been requested). Descendent classes extend this
        method to finalize encoder configuration.

        .. note::

            It should be noted that this method is called with the
            initializer's ``option`` keyword arguments. This base
            implementation expects no additional arguments, but descendent
            classes extend the parameter list to include options relevant to
            them.
        """
        assert not self.encoder
        self.encoder = self.encoder_type()
        self.output_port = self.encoder.outputs[0]
        if self.resizer:
            self.encoder.inputs[0].connect(self.resizer.outputs[0])
        else:
            self.encoder.inputs[0].connect(self.input_port)
        self.encoder.outputs[0].copy_from(self.encoder.inputs[0])
        # NOTE: We deliberately don't commit the output port format here as
        # this is a base class and the output configuration is incomplete at
        # this point. Descendents are expected to finish configuring the
        # encoder and then commit the port format themselves

    def _callback(self, port, buf):
        """
        The encoder's main callback function.

        When the encoder is active, this method is periodically called in a
        background thread. The *port* parameter specifies the :class:`MMALPort`
        providing the output (typically this is the encoder's output port, but
        in the case of unencoded captures may simply be a camera port), while
        the *buf* parameter is an :class:`~mmalobj.MMALBuffer` which can be
        used to obtain the data to write, along with meta-data about the
        current frame.

        This method must set :attr:`event` when the encoder has finished (and
        should set :attr:`exception` if an exception occurred during encoding).

        Developers wishing to write a custom encoder class may find it simpler
        to override the :meth:`_callback_write` method, rather than deal with
        these complexities.
        """
        if self.DEBUG > 1:
            print(repr(buf))
        try:
            stop = self._callback_write(buf)
        except Exception as e:
            stop = True
            self.exception = e
        if stop:
            self.event.set()
        return stop

    def _callback_write(self, buf, key=PiVideoFrameType.frame):
        """
        _callback_write(buf, key=PiVideoFrameType.frame)

        Writes output on behalf of the encoder callback function.

        This method is called by :meth:`_callback` to handle writing to an
        object in :attr:`outputs` identified by *key*. The *buf* parameter is
        an :class:`~mmalobj.MMALBuffer` which can be used to obtain the data.
        The method is expected to return a boolean to indicate whether output
        is complete (``True``) or whether more data is expected (``False``).

        The default implementation simply writes the contents of the buffer to
        the output identified by *key*, and returns ``True`` if the buffer
        flags indicate end of stream. Image encoders will typically override
        the return value to indicate ``True`` on end of frame (as they only
        wish to output a single image). Video encoders will typically override
        this method to determine where key-frames and SPS headers occur.
        """
        if buf.length:
            with self.outputs_lock:
                try:
                    output = self.outputs[key][0]
                    written = output.write(buf.data)
                except KeyError:
                    # No output associated with the key type; discard the
                    # data
                    pass
                else:
                    # Ignore None return value; most Python 2 streams have
                    # no return value for write()
                    if (written is not None) and (written != buf.length):
                        raise PiCameraIOError(
                            "Failed to write %d bytes from buffer to "
                            "output %r" % (buf.length, output))
        return bool(buf.flags & mmal.MMAL_BUFFER_HEADER_FLAG_EOS)

    def _open_output(self, output, key=PiVideoFrameType.frame):
        """
        _open_output(output, key=PiVideoFrameType.frame)

        Opens *output* and associates it with *key* in :attr:`outputs`.

        If *output* is a string, this method opens it as a filename and keeps
        track of the fact that the encoder was the one to open it (which
        implies that :meth:`_close_output` should eventually close it).
        Otherwise, if *output* has a ``write`` method it is assumed to be a
        file-like object and it is used verbatim. If *output* is neither a
        string, nor an object with a ``write`` method it is assumed to be a
        writeable object supporting the buffer protocol (this is wrapped in
        a :class:`BufferIO` stream to simplify writing).

        The opened output is added to the :attr:`outputs` dictionary with the
        specified *key*.
        """
        with self.outputs_lock:
            self.outputs[key] = mo.open_stream(output)

    def _close_output(self, key=PiVideoFrameType.frame):
        """
        _close_output(key=PiVideoFrameType.frame)

        Closes the output associated with *key* in :attr:`outputs`.

        Closes the output object associated with the specified *key*, and
        removes it from the :attr:`outputs` dictionary (if we didn't open the
        object then we attempt to flush it instead).
        """
        with self.outputs_lock:
            try:
                (output, opened) = self.outputs.pop(key)
            except KeyError:
                pass
            else:
                mo.close_stream(output, opened)

    @property
    def active(self):
        """
        Returns ``True`` if the MMAL encoder exists and is enabled.
        """
        try:
            return bool(self.output_port.enabled)
        except AttributeError:
            # output_port can be None; avoid a (demonstrated) race condition
            # by catching AttributeError
            return False

    def start(self, output):
        """
        Starts the encoder object writing to the specified output.

        This method is called by the camera to start the encoder capturing
        data from the camera to the specified output. The *output* parameter
        is either a filename, or a file-like object (for image and video
        encoders), or an iterable of filenames or file-like objects (for
        multi-image encoders).
        """
        self.event.clear()
        self.exception = None
        self._open_output(output)
        with self.parent._encoders_lock:
            self.output_port.enable(self._callback)
            if self.DEBUG > 0:
                mo.print_pipeline(self.output_port)
            self.parent._start_capture(self.camera_port)

    def wait(self, timeout=None):
        """
        Waits for the encoder to finish (successfully or otherwise).

        This method is called by the owning camera object to block execution
        until the encoder has completed its task. If the *timeout* parameter
        is None, the method will block indefinitely. Otherwise, the *timeout*
        parameter specifies the (potentially fractional) number of seconds
        to block for. If the encoder finishes successfully within the timeout,
        the method returns ``True``. Otherwise, it returns ``False``.
        """
        result = self.event.wait(timeout)
        if result:
            self.stop()
            # Check whether the callback set an exception
            if self.exception:
                raise self.exception
        return result

    def stop(self):
        """
        Stops the encoder, regardless of whether it's finished.

        This method is called by the camera to terminate the execution of the
        encoder. Typically, this is used with video to stop the recording, but
        can potentially be called in the middle of image capture to terminate
        the capture.
        """
        # NOTE: The active test below is necessary to prevent attempting to
        # re-enter the parent lock in the case the encoder is being torn down
        # by an error in the constructor
        if self.active:
            if self.parent and self.camera_port:
                with self.parent._encoders_lock:
                    self.parent._stop_capture(self.camera_port)
            self.output_port.disable()
        self.event.set()
        self._close_output()

    def close(self):
        """
        Finalizes the encoder and deallocates all structures.

        This method is called by the camera prior to destroying the encoder (or
        more precisely, letting it go out of scope to permit the garbage
        collector to destroy it at some future time). The method destroys all
        components that the various create methods constructed and resets their
        attributes.
        """
        self.stop()
        if self.encoder:
            self.encoder.disconnect()
        if self.resizer:
            self.resizer.disconnect()
        if self.encoder:
            self.encoder.close()
            self.encoder = None
        if self.resizer:
            self.resizer.close()
            self.resizer = None
        self.output_port = None


class MMALBufferAlphaStrip(mo.MMALBuffer):
    """
    An MMALBuffer descendent that strips alpha bytes from the buffer data. This
    is used internally by PiRawMixin when it needs to strip alpha bytes itself
    (e.g. because an appropriate format cannot be selected on an output port).
    """

    def __init__(self, buf):
        super(MMALBufferAlphaStrip, self).__init__(buf)
        self._stripped = bytearray(super(MMALBufferAlphaStrip, self).data)
        del self._stripped[3::4]

    @property
    def length(self):
        return len(self._stripped)

    @property
    def data(self):
        return self._stripped


class PiRawMixin(PiEncoder):
    """
    Mixin class for "raw" (unencoded) output.

    This mixin class overrides the initializer of :class:`PiEncoder`, along
    with :meth:`_create_resizer` and :meth:`_create_encoder` to configure the
    pipeline for unencoded output. Specifically, it disables the construction
    of an encoder, and sets the output port to the input port passed to the
    initializer, unless resizing is required (either for actual resizing, or
    for format conversion) in which case the resizer's output is used.
    """

    RAW_ENCODINGS = {
        # name   mmal-encoding             bytes-per-pixel
        'yuv':  (mmal.MMAL_ENCODING_I420,  1.5),
        'rgb':  (mmal.MMAL_ENCODING_RGB24, 3),
        'rgba': (mmal.MMAL_ENCODING_RGBA,  4),
        'bgr':  (mmal.MMAL_ENCODING_BGR24, 3),
        'bgra': (mmal.MMAL_ENCODING_BGRA,  4),
        }

    def __init__(
            self, parent, camera_port, input_port, format, resize, **options):
        encoding, bpp = self.RAW_ENCODINGS[format]
        # Workaround: on older firmwares, non-YUV encodings aren't supported on
        # the still port. If a non-YUV format is requested without resizing,
        # test whether we can commit the requested format on the input port and
        # if this fails, set resize to force resizer usage
        if resize is None and encoding != mmal.MMAL_ENCODING_I420:
            input_port.format = encoding
            try:
                input_port.commit()
            except PiCameraMMALError as e:
                if e.status != mmal.MMAL_EINVAL:
                    raise
                resize = input_port.framesize
                warnings.warn(
                    PiCameraResizerEncoding(
                        "using a resizer to perform non-YUV encoding; "
                        "upgrading your firmware with sudo rpi-update "
                        "may improve performance"))
        # Workaround: If a non-alpha format is requested with the resizer, use
        # the alpha-inclusive format and set a flag to get the callback to
        # strip the alpha bytes
        self._strip_alpha = False
        if resize:
            width, height = resize
            try:
                format = {
                    'rgb': 'rgba',
                    'bgr': 'bgra',
                    }[format]
                self._strip_alpha = True
                warnings.warn(
                    PiCameraAlphaStripping(
                        "using alpha-stripping to convert to non-alpha "
                        "format; you may find the equivalent alpha format "
                        "faster"))
            except KeyError:
                pass
        else:
            width, height = input_port.framesize
        # Workaround (#83): when the resizer is used the width must be aligned
        # (both the frame and crop values) to avoid an error when the output
        # port format is set (height is aligned too, simply for consistency
        # with old picamera versions). Warn the user as they're not going to
        # get the resolution they expect
        if not resize and format != 'yuv' and input_port.name.startswith('vc.ril.video_splitter'):
            # Workaround: Expected frame size is rounded to 16x16 when splitter
            # port with no resizer is used and format is not YUV
            fwidth = bcm_host.VCOS_ALIGN_UP(width, 16)
        else:
            fwidth = bcm_host.VCOS_ALIGN_UP(width, 32)
        fheight = bcm_host.VCOS_ALIGN_UP(height, 16)
        if fwidth != width or fheight != height:
            warnings.warn(
                PiCameraResolutionRounded(
                    "frame size rounded up from %dx%d to %dx%d" % (
                        width, height, fwidth, fheight)))
        if resize:
            resize = (fwidth, fheight)
        # Workaround: Calculate the expected frame size, to be used by the
        # callback to decide when a frame ends. This is to work around a
        # firmware bug that causes the raw image to be returned twice when the
        # maximum camera resolution is requested
        self._frame_size = int(fwidth * fheight * bpp)
        super(PiRawMixin, self).__init__(
                parent, camera_port, input_port, format, resize, **options)

    def _create_encoder(self, format):
        """
        Overridden to skip creating an encoder. Instead, this class simply uses
        the resizer's port as the output port (if a resizer has been
        configured) or the specified input port otherwise.
        """
        if self.resizer:
            self.output_port = self.resizer.outputs[0]
        else:
            self.output_port = self.input_port
        try:
            self.output_port.format = self.RAW_ENCODINGS[format][0]
        except KeyError:
            raise PiCameraValueError('unknown format %s' % format)
        self.output_port.commit()

    def _callback_write(self, buf, key=PiVideoFrameType.frame):
        """
        _callback_write(buf, key=PiVideoFrameType.frame)

        Overridden to strip alpha bytes when required.
        """
        if self._strip_alpha:
            return super(PiRawMixin, self)._callback_write(MMALBufferAlphaStrip(buf._buf), key)
        else:
            return super(PiRawMixin, self)._callback_write(buf, key)


class PiVideoEncoder(PiEncoder):
    """
    Encoder for video recording.

    This derivative of :class:`PiEncoder` configures itself for H.264 or MJPEG
    encoding.  It also introduces a :meth:`split` method which is used by
    :meth:`~PiCamera.split_recording` and :meth:`~PiCamera.record_sequence` to
    redirect future output to a new filename or object. Finally, it also
    extends :meth:`PiEncoder.start` and :meth:`PiEncoder._callback_write` to
    track video frame meta-data, and to permit recording motion data to a
    separate output object.
    """

    encoder_type = mo.MMALVideoEncoder

    def __init__(
            self, parent, camera_port, input_port, format, resize, **options):
        super(PiVideoEncoder, self).__init__(
                parent, camera_port, input_port, format, resize, **options)
        self._next_output = []
        self.frame = None

    def _create_encoder(
            self, format, bitrate=17000000, intra_period=None, profile='high',
            level='4', quantization=0, quality=0, inline_headers=True,
            sei=False, sps_timing=False, motion_output=None,
            intra_refresh=None):
        """
        Extends the base :meth:`~PiEncoder._create_encoder` implementation to
        configure the video encoder for H.264 or MJPEG output.
        """
        super(PiVideoEncoder, self)._create_encoder(format)

        # XXX Remove quantization in 2.0
        quality = quality or quantization

        try:
            self.output_port.format = {
                'h264':  mmal.MMAL_ENCODING_H264,
                'mjpeg': mmal.MMAL_ENCODING_MJPEG,
                }[format]
        except KeyError:
            raise PiCameraValueError('Unsupported format %s' % format)

        if format == 'h264':
            try:
                profile = {
                    'baseline':    mmal.MMAL_VIDEO_PROFILE_H264_BASELINE,
                    'main':        mmal.MMAL_VIDEO_PROFILE_H264_MAIN,
                    'extended':    mmal.MMAL_VIDEO_PROFILE_H264_EXTENDED,
                    'high':        mmal.MMAL_VIDEO_PROFILE_H264_HIGH,
                    'constrained': mmal.MMAL_VIDEO_PROFILE_H264_CONSTRAINED_BASELINE,
                    }[profile]
            except KeyError:
                raise PiCameraValueError("Invalid H.264 profile %s" % profile)
            try:
                level = {
                    '1':   mmal.MMAL_VIDEO_LEVEL_H264_1,
                    '1.0': mmal.MMAL_VIDEO_LEVEL_H264_1,
                    '1b':  mmal.MMAL_VIDEO_LEVEL_H264_1b,
                    '1.1': mmal.MMAL_VIDEO_LEVEL_H264_11,
                    '1.2': mmal.MMAL_VIDEO_LEVEL_H264_12,
                    '1.3': mmal.MMAL_VIDEO_LEVEL_H264_13,
                    '2':   mmal.MMAL_VIDEO_LEVEL_H264_2,
                    '2.0': mmal.MMAL_VIDEO_LEVEL_H264_2,
                    '2.1': mmal.MMAL_VIDEO_LEVEL_H264_21,
                    '2.2': mmal.MMAL_VIDEO_LEVEL_H264_22,
                    '3':   mmal.MMAL_VIDEO_LEVEL_H264_3,
                    '3.0': mmal.MMAL_VIDEO_LEVEL_H264_3,
                    '3.1': mmal.MMAL_VIDEO_LEVEL_H264_31,
                    '3.2': mmal.MMAL_VIDEO_LEVEL_H264_32,
                    '4':   mmal.MMAL_VIDEO_LEVEL_H264_4,
                    '4.0': mmal.MMAL_VIDEO_LEVEL_H264_4,
                    '4.1': mmal.MMAL_VIDEO_LEVEL_H264_41,
                    '4.2': mmal.MMAL_VIDEO_LEVEL_H264_42,
                    }[level]
            except KeyError:
                raise PiCameraValueError("Invalid H.264 level %s" % level)

            # From https://en.wikipedia.org/wiki/H.264/MPEG-4_AVC#Levels
            bitrate_limit = {
                # level, high-profile:  bitrate
                (mmal.MMAL_VIDEO_LEVEL_H264_1,  False): 64000,
                (mmal.MMAL_VIDEO_LEVEL_H264_1b, False): 128000,
                (mmal.MMAL_VIDEO_LEVEL_H264_11, False): 192000,
                (mmal.MMAL_VIDEO_LEVEL_H264_12, False): 384000,
                (mmal.MMAL_VIDEO_LEVEL_H264_13, False): 768000,
                (mmal.MMAL_VIDEO_LEVEL_H264_2,  False): 2000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_21, False): 4000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_22, False): 4000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_3,  False): 10000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_31, False): 14000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_32, False): 20000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_4,  False): 20000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_41, False): 50000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_42, False): 50000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_1,  True):  80000,
                (mmal.MMAL_VIDEO_LEVEL_H264_1b, True):  160000,
                (mmal.MMAL_VIDEO_LEVEL_H264_11, True):  240000,
                (mmal.MMAL_VIDEO_LEVEL_H264_12, True):  480000,
                (mmal.MMAL_VIDEO_LEVEL_H264_13, True):  960000,
                (mmal.MMAL_VIDEO_LEVEL_H264_2,  True):  2500000,
                (mmal.MMAL_VIDEO_LEVEL_H264_21, True):  5000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_22, True):  5000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_3,  True):  12500000,
                (mmal.MMAL_VIDEO_LEVEL_H264_31, True):  17500000,
                (mmal.MMAL_VIDEO_LEVEL_H264_32, True):  25000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_4,  True):  25000000,
                (mmal.MMAL_VIDEO_LEVEL_H264_41, True):  62500000,
                (mmal.MMAL_VIDEO_LEVEL_H264_42, True):  62500000,
                }[level, profile == mmal.MMAL_VIDEO_PROFILE_H264_HIGH]
            if bitrate > bitrate_limit:
                raise PiCameraValueError(
                    'bitrate %d exceeds %d which is the limit for the '
                    'selected H.264 level and profile' %
                    (bitrate, bitrate_limit))
            self.output_port.bitrate = bitrate
            self.output_port.commit()

            # Again, from https://en.wikipedia.org/wiki/H.264/MPEG-4_AVC#Levels
            macroblocks_per_s_limit, macroblocks_limit = {
                #level: macroblocks/s, macroblocks
                mmal.MMAL_VIDEO_LEVEL_H264_1:  (1485,   99),
                mmal.MMAL_VIDEO_LEVEL_H264_1b: (1485,   99),
                mmal.MMAL_VIDEO_LEVEL_H264_11: (3000,   396),
                mmal.MMAL_VIDEO_LEVEL_H264_12: (6000,   396),
                mmal.MMAL_VIDEO_LEVEL_H264_13: (11880,  396),
                mmal.MMAL_VIDEO_LEVEL_H264_2:  (11880,  396),
                mmal.MMAL_VIDEO_LEVEL_H264_21: (19800,  792),
                mmal.MMAL_VIDEO_LEVEL_H264_22: (20250,  1620),
                mmal.MMAL_VIDEO_LEVEL_H264_3:  (40500,  1620),
                mmal.MMAL_VIDEO_LEVEL_H264_31: (108000, 3600),
                mmal.MMAL_VIDEO_LEVEL_H264_32: (216000, 5120),
                mmal.MMAL_VIDEO_LEVEL_H264_4:  (245760, 8192),
                mmal.MMAL_VIDEO_LEVEL_H264_41: (245760, 8192),
                mmal.MMAL_VIDEO_LEVEL_H264_42: (522240, 8704),
                }[level]
            w, h = self.output_port.framesize
            w = bcm_host.VCOS_ALIGN_UP(w, 16) >> 4
            h = bcm_host.VCOS_ALIGN_UP(h, 16) >> 4
            if w * h > macroblocks_limit:
                raise PiCameraValueError(
                    'output resolution %s exceeds macroblock limit (%d) for '
                    'the selected H.264 profile and level' %
                    (self.output_port.framesize, macroblocks_limit))
            if self.parent:
                framerate = self.parent.framerate + self.parent.framerate_delta
            else:
                framerate = self.input_port.framerate
            if w * h * framerate > macroblocks_per_s_limit:
                raise PiCameraValueError(
                    'output resolution and framerate exceeds macroblocks/s '
                    'limit (%d) for the selected H.264 profile and '
                    'level' % macroblocks_per_s_limit)

            mp = mmal.MMAL_PARAMETER_VIDEO_PROFILE_T(
                    mmal.MMAL_PARAMETER_HEADER_T(
                        mmal.MMAL_PARAMETER_PROFILE,
                        ct.sizeof(mmal.MMAL_PARAMETER_VIDEO_PROFILE_T),
                        ),
                    )
            mp.profile[0].profile = profile
            mp.profile[0].level = level
            self.output_port.params[mmal.MMAL_PARAMETER_PROFILE] = mp

            if inline_headers:
                self.output_port.params[mmal.MMAL_PARAMETER_VIDEO_ENCODE_INLINE_HEADER] = True
            if sei:
                self.output_port.params[mmal.MMAL_PARAMETER_VIDEO_ENCODE_SEI_ENABLE] = True
            if sps_timing:
                self.output_port.params[mmal.MMAL_PARAMETER_VIDEO_ENCODE_SPS_TIMING] = True
            if motion_output is not None:
                self.output_port.params[mmal.MMAL_PARAMETER_VIDEO_ENCODE_INLINE_VECTORS] = True

            # We need the intra-period to calculate the SPS header timeout in
            # the split method below. If one is not set explicitly, query the
            # encoder's default
            if intra_period is not None:
                self.output_port.params[mmal.MMAL_PARAMETER_INTRAPERIOD] = intra_period
                self._intra_period = intra_period
            else:
                self._intra_period = self.output_port.params[mmal.MMAL_PARAMETER_INTRAPERIOD]

            if intra_refresh is not None:
                # Get the intra-refresh structure first as there are several
                # other fields in it which we don't wish to overwrite
                mp = self.output_port.params[mmal.MMAL_PARAMETER_VIDEO_INTRA_REFRESH]
                try:
                    mp.refresh_mode = {
                        'cyclic':     mmal.MMAL_VIDEO_INTRA_REFRESH_CYCLIC,
                        'adaptive':   mmal.MMAL_VIDEO_INTRA_REFRESH_ADAPTIVE,
                        'both':       mmal.MMAL_VIDEO_INTRA_REFRESH_BOTH,
                        'cyclicrows': mmal.MMAL_VIDEO_INTRA_REFRESH_CYCLIC_MROWS,
                        }[intra_refresh]
                except KeyError:
                    raise PiCameraValueError(
                        "Invalid intra_refresh %s" % intra_refresh)
                self.output_port.params[mmal.MMAL_PARAMETER_VIDEO_INTRA_REFRESH] = mp

        elif format == 'mjpeg':
            self.output_port.bitrate = bitrate
            self.output_port.commit()
            # MJPEG doesn't have an intra_period setting as such, but as every
            # frame is a full-frame, the intra_period is effectively 1
            self._intra_period = 1

        if quality:
            self.output_port.params[mmal.MMAL_PARAMETER_VIDEO_ENCODE_INITIAL_QUANT] = quality
            self.output_port.params[mmal.MMAL_PARAMETER_VIDEO_ENCODE_MIN_QUANT] = quality
            self.output_port.params[mmal.MMAL_PARAMETER_VIDEO_ENCODE_MAX_QUANT] = quality

        self.encoder.inputs[0].params[mmal.MMAL_PARAMETER_VIDEO_IMMUTABLE_INPUT] = True
        self.encoder.enable()

    def start(self, output, motion_output=None):
        """
        Extended to initialize video frame meta-data tracking.
        """
        self.frame = PiVideoFrame(
                index=0,
                frame_type=None,
                frame_size=0,
                video_size=0,
                split_size=0,
                timestamp=0,
                complete=False,
                )
        if motion_output is not None:
            self._open_output(motion_output, PiVideoFrameType.motion_data)
        super(PiVideoEncoder, self).start(output)

    def stop(self):
        super(PiVideoEncoder, self).stop()
        self._close_output(PiVideoFrameType.motion_data)

    def request_key_frame(self):
        """
        Called to request an I-frame from the encoder.

        This method is called by :meth:`~PiCamera.request_key_frame` and
        :meth:`split` to force the encoder to output an I-frame as soon as
        possible.
        """
        self.encoder.control.params[mmal.MMAL_PARAMETER_VIDEO_REQUEST_I_FRAME] = True

    def split(self, output, motion_output=None):
        """
        Called to switch the encoder's output.

        This method is called by :meth:`~PiCamera.split_recording` and
        :meth:`~PiCamera.record_sequence` to switch the encoder's
        :attr:`output` object to the *output* parameter (which can be a
        filename or a file-like object, as with :meth:`start`).
        """
        with self.outputs_lock:
            outputs = {}
            if output is not None:
                outputs[PiVideoFrameType.frame] = output
            if motion_output is not None:
                outputs[PiVideoFrameType.motion_data] = motion_output
            self._next_output.append(outputs)
        # intra_period / framerate gives the time between I-frames (which
        # should also coincide with SPS headers). We multiply by three to
        # ensure the timeout is deliberately excessive, and clamp the minimum
        # timeout to 10 seconds (otherwise unencoded formats tend to fail
        # presumably due to I/O capacity)
        if self.parent:
            framerate = self.parent.framerate + self.parent.framerate_delta
        else:
            framerate = self.input_port.framerate
        timeout = max(15.0, float(self._intra_period / framerate) * 3.0)
        if self._intra_period > 1:
            self.request_key_frame()
        if not self.event.wait(timeout):
            raise PiCameraRuntimeError('Timed out waiting for a split point')
        self.event.clear()

    def _callback_write(self, buf, key=PiVideoFrameType.frame):
        """
        Extended to implement video frame meta-data tracking, and to handle
        splitting video recording to the next output when :meth:`split` is
        called.
        """
        self.frame = PiVideoFrame(
            index=
                self.frame.index + 1
                if self.frame.complete else
                self.frame.index,
            frame_type=
                PiVideoFrameType.key_frame
                if buf.flags & mmal.MMAL_BUFFER_HEADER_FLAG_KEYFRAME else
                PiVideoFrameType.sps_header
                if buf.flags & mmal.MMAL_BUFFER_HEADER_FLAG_CONFIG else
                PiVideoFrameType.motion_data
                if buf.flags & mmal.MMAL_BUFFER_HEADER_FLAG_CODECSIDEINFO else
                PiVideoFrameType.frame,
            frame_size=
                buf.length
                if self.frame.complete else
                self.frame.frame_size + buf.length,
            video_size=
                self.frame.video_size
                if buf.flags & mmal.MMAL_BUFFER_HEADER_FLAG_CODECSIDEINFO else
                self.frame.video_size + buf.length,
            split_size=
                self.frame.split_size
                if buf.flags & mmal.MMAL_BUFFER_HEADER_FLAG_CODECSIDEINFO else
                self.frame.split_size + buf.length,
            timestamp=
                None
                if buf.pts in (0, mmal.MMAL_TIME_UNKNOWN) else
                buf.pts,
            complete=
                bool(buf.flags & mmal.MMAL_BUFFER_HEADER_FLAG_FRAME_END),
            )
        if self._intra_period == 1 or (buf.flags & mmal.MMAL_BUFFER_HEADER_FLAG_CONFIG):
            with self.outputs_lock:
                try:
                    new_outputs = self._next_output.pop(0)
                except IndexError:
                    new_outputs = None
            if new_outputs:
                for new_key, new_output in new_outputs.items():
                    self._close_output(new_key)
                    self._open_output(new_output, new_key)
                    if new_key == PiVideoFrameType.frame:
                        self.frame = PiVideoFrame(
                                index=self.frame.index,
                                frame_type=self.frame.frame_type,
                                frame_size=self.frame.frame_size,
                                video_size=self.frame.video_size,
                                split_size=0,
                                timestamp=self.frame.timestamp,
                                complete=self.frame.complete,
                                )
                self.event.set()
        if buf.flags & mmal.MMAL_BUFFER_HEADER_FLAG_CODECSIDEINFO:
            key = PiVideoFrameType.motion_data
        return super(PiVideoEncoder, self)._callback_write(buf, key)


class PiCookedVideoEncoder(PiVideoEncoder):
    """
    Video encoder for encoded recordings.

    This class is a derivative of :class:`PiVideoEncoder` and only exists to
    provide naming symmetry with the image encoder classes.
    """


class PiRawVideoEncoder(PiRawMixin, PiVideoEncoder):
    """
    Video encoder for unencoded recordings.

    This class is a derivative of :class:`PiVideoEncoder` and the
    :class:`PiRawMixin` class intended for use with
    :meth:`~PiCamera.start_recording` when it is called with an unencoded
    format.

    .. warning::

        This class creates an inheritance diamond. Take care to determine the
        MRO of super-class calls.
    """

    def _create_encoder(self, format):
        super(PiRawVideoEncoder, self)._create_encoder(format)
        # Raw formats don't have an intra_period setting as such, but as every
        # frame is a full-frame, the intra_period is effectively 1
        self._intra_period = 1


class PiImageEncoder(PiEncoder):
    """
    Encoder for image capture.

    This derivative of :class:`PiEncoder` extends the :meth:`_create_encoder`
    method to configure the encoder for a variety of encoded image outputs
    (JPEG, PNG, etc.).
    """

    encoder_type = mo.MMALImageEncoder

    def _create_encoder(
            self, format, quality=85, thumbnail=(64, 48, 35), restart=0):
        """
        Extends the base :meth:`~PiEncoder._create_encoder` implementation to
        configure the image encoder for JPEG, PNG, etc.
        """
        super(PiImageEncoder, self)._create_encoder(format)

        try:
            self.output_port.format = {
                'jpeg': mmal.MMAL_ENCODING_JPEG,
                'png':  mmal.MMAL_ENCODING_PNG,
                'gif':  mmal.MMAL_ENCODING_GIF,
                'bmp':  mmal.MMAL_ENCODING_BMP,
                }[format]
        except KeyError:
            raise PiCameraValueError("Unsupported format %s" % format)
        self.output_port.commit()

        if format == 'jpeg':
            self.output_port.params[mmal.MMAL_PARAMETER_JPEG_Q_FACTOR] = quality
            if restart > 0:
                # Don't set if zero as old firmwares don't support this param
                self.output_port.params[mmal.MMAL_PARAMETER_JPEG_RESTART_INTERVAL] = restart
            if thumbnail is None:
                mp = mmal.MMAL_PARAMETER_THUMBNAIL_CONFIG_T(
                    mmal.MMAL_PARAMETER_HEADER_T(
                        mmal.MMAL_PARAMETER_THUMBNAIL_CONFIGURATION,
                        ct.sizeof(mmal.MMAL_PARAMETER_THUMBNAIL_CONFIG_T)
                        ),
                    0, 0, 0, 0)
            else:
                mp = mmal.MMAL_PARAMETER_THUMBNAIL_CONFIG_T(
                    mmal.MMAL_PARAMETER_HEADER_T(
                        mmal.MMAL_PARAMETER_THUMBNAIL_CONFIGURATION,
                        ct.sizeof(mmal.MMAL_PARAMETER_THUMBNAIL_CONFIG_T)
                        ),
                    1, *thumbnail)
            self.encoder.control.params[mmal.MMAL_PARAMETER_THUMBNAIL_CONFIGURATION] = mp

        self.encoder.enable()


class PiOneImageEncoder(PiImageEncoder):
    """
    Encoder for single image capture.

    This class simply extends :meth:`~PiEncoder._callback_write` to terminate
    capture at frame end (i.e. after a single frame has been received).
    """

    def _callback_write(self, buf, key=PiVideoFrameType.frame):
        return (
            super(PiOneImageEncoder, self)._callback_write(buf, key)
            ) or bool(
            buf.flags & (
                mmal.MMAL_BUFFER_HEADER_FLAG_FRAME_END |
                mmal.MMAL_BUFFER_HEADER_FLAG_TRANSMISSION_FAILED)
            )


class PiMultiImageEncoder(PiImageEncoder):
    """
    Encoder for multiple image capture.

    This class extends :class:`PiImageEncoder` to handle an iterable of outputs
    instead of a single output. The :meth:`~PiEncoder._callback_write` method
    is extended to terminate capture when the iterable is exhausted, while
    :meth:`PiEncoder._open_output` is overridden to begin iteration and rely
    on the new :meth:`_next_output` method to advance output to the next item
    in the iterable.
    """

    def _open_output(self, outputs, key=PiVideoFrameType.frame):
        self._output_iter = iter(outputs)
        self._next_output(key)

    def _next_output(self, key=PiVideoFrameType.frame):
        """
        This method moves output to the next item from the iterable passed to
        :meth:`~PiEncoder.start`.
        """
        self._close_output(key)
        super(PiMultiImageEncoder, self)._open_output(next(self._output_iter), key)

    def _callback_write(self, buf, key=PiVideoFrameType.frame):
        try:
            if (
                super(PiMultiImageEncoder, self)._callback_write(buf, key)
                ) or bool(
                buf.flags & (
                    mmal.MMAL_BUFFER_HEADER_FLAG_FRAME_END |
                    mmal.MMAL_BUFFER_HEADER_FLAG_TRANSMISSION_FAILED)
                ):
                self._next_output(key)
            return False
        except StopIteration:
            return True


class PiCookedOneImageEncoder(PiOneImageEncoder):
    """
    Encoder for "cooked" (encoded) single image output.

    This encoder extends :class:`PiOneImageEncoder` to include Exif tags in the
    output.
    """

    exif_encoding = 'ascii'

    def __init__(
            self, parent, camera_port, input_port, format, resize, **options):
        super(PiCookedOneImageEncoder, self).__init__(
                parent, camera_port, input_port, format, resize, **options)
        if parent:
            self.exif_tags = self.parent.exif_tags
        else:
            self.exif_tags = {}

    def _add_exif_tag(self, tag, value):
        # Format the tag and value into an appropriate bytes string, encoded
        # with the Exif encoding (ASCII)
        if isinstance(tag, str):
            tag = tag.encode(self.exif_encoding)
        if isinstance(value, str):
            value = value.encode(self.exif_encoding)
        elif isinstance(value, datetime.datetime):
            value = value.strftime('%Y:%m:%d %H:%M:%S').encode(self.exif_encoding)
        # MMAL_PARAMETER_EXIF_T is a variable sized structure, hence all the
        # mucking about with string buffers here...
        buf = ct.create_string_buffer(
            ct.sizeof(mmal.MMAL_PARAMETER_EXIF_T) + len(tag) + len(value) + 1)
        mp = ct.cast(buf, ct.POINTER(mmal.MMAL_PARAMETER_EXIF_T))
        mp[0].hdr.id = mmal.MMAL_PARAMETER_EXIF
        mp[0].hdr.size = len(buf)
        if (b'=' in tag or b'\x00' in value):
            data = tag + value
            mp[0].keylen = len(tag)
            mp[0].value_offset = len(tag)
            mp[0].valuelen = len(value)
        else:
            data = tag + b'=' + value
        ct.memmove(mp[0].data, data, len(data))
        self.output_port.params[mmal.MMAL_PARAMETER_EXIF] = mp[0]

    def start(self, output):
        timestamp = datetime.datetime.now()
        timestamp_tags = (
            'EXIF.DateTimeDigitized',
            'EXIF.DateTimeOriginal',
            'IFD0.DateTime')
        # Timestamp tags are always included with the value calculated
        # above, but the user may choose to override the value in the
        # exif_tags mapping
        for tag in timestamp_tags:
            self._add_exif_tag(tag, self.exif_tags.get(tag, timestamp))
        # All other tags are just copied in verbatim
        for tag, value in self.exif_tags.items():
            if not tag in timestamp_tags:
                self._add_exif_tag(tag, value)
        super(PiCookedOneImageEncoder, self).start(output)


class PiCookedMultiImageEncoder(PiMultiImageEncoder):
    """
    Encoder for "cooked" (encoded) multiple image output.

    This encoder descends from :class:`PiMultiImageEncoder` but includes no
    new functionality as video-port based encodes (which is all this class
    is used for) don't support Exif tag output.
    """
    pass


class PiRawImageMixin(PiRawMixin, PiImageEncoder):
    """
    Mixin class for "raw" (unencoded) image capture.

    The :meth:`_callback_write` method is overridden to manually calculate when
    to terminate output.
    """

    def __init__(
            self, parent, camera_port, input_port, format, resize, **options):
        super(PiRawImageMixin, self).__init__(
                parent, camera_port, input_port, format, resize, **options)
        self._image_size = 0

    def _callback_write(self, buf, key=PiVideoFrameType.frame):
        """
        Overridden to manually calculate when to terminate capture (see
        comments in :meth:`__init__`).
        """
        if self._image_size > 0:
            super(PiRawImageMixin, self)._callback_write(buf, key)
            self._image_size -= buf.length
        return self._image_size <= 0

    def start(self, output):
        self._image_size = self._frame_size
        super(PiRawImageMixin, self).start(output)


class PiRawOneImageEncoder(PiOneImageEncoder, PiRawImageMixin):
    """
    Single image encoder for unencoded capture.

    This class is a derivative of :class:`PiOneImageEncoder` and the
    :class:`PiRawImageMixin` class intended for use with
    :meth:`~PiCamera.capture` (et al) when it is called with an unencoded image
    format.

    .. warning::

        This class creates an inheritance diamond. Take care to determine the
        MRO of super-class calls.
    """
    pass


class PiRawMultiImageEncoder(PiMultiImageEncoder, PiRawImageMixin):
    """
    Multiple image encoder for unencoded capture.

    This class is a derivative of :class:`PiMultiImageEncoder` and the
    :class:`PiRawImageMixin` class intended for use with
    :meth:`~PiCamera.capture_sequence` when it is called with an unencoded
    image format.

    .. warning::

        This class creates an inheritance diamond. Take care to determine the
        MRO of super-class calls.
    """
    def _next_output(self, key=PiVideoFrameType.frame):
        super(PiRawMultiImageEncoder, self)._next_output(key)
        self._image_size = self._frame_size




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Python header conversion
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Original headers
# Copyright (c) 2012, Broadcom Europe Ltd
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    print_function,
    division,
    absolute_import,
    )

# Make Py2's str equivalent to Py3's
str = type('')

import ctypes as ct
import warnings

_lib = ct.CDLL('libbcm_host.so')

# bcm_host.h #################################################################

bcm_host_init = _lib.bcm_host_init
bcm_host_init.argtypes = []
bcm_host_init.restype = None

bcm_host_deinit = _lib.bcm_host_deinit
bcm_host_deinit.argtypes = []
bcm_host_deinit.restype = None

graphics_get_display_size = _lib.graphics_get_display_size
graphics_get_display_size.argtypes = [ct.c_uint16, ct.POINTER(ct.c_uint32), ct.POINTER(ct.c_uint32)]
graphics_get_display_size.restype = ct.c_int32

# vchi.h #####################################################################

VCHI_INSTANCE_T = ct.c_void_p
VCHI_CONNECTION_T = ct.c_void_p

# vcos_platform.h ############################################################

VCOS_UNSIGNED = ct.c_uint32

# vcos_types.h ###############################################################

VCOS_STATUS_T = ct.c_uint32 # enum
(
    VCOS_SUCCESS,
    VCOS_EAGAIN,
    VCOS_ENOENT,
    VCOS_ENOSPC,
    VCOS_EINVAL,
    VCOS_EACCESS,
    VCOS_ENOMEM,
    VCOS_ENOSYS,
    VCOS_EEXIST,
    VCOS_ENXIO,
    VCOS_EINTR,
) = range(11)

vcos_bool_t = ct.c_int32
vcos_fourcc_t = ct.c_int32

def VCOS_ALIGN_UP(value, round_to):
    # Note: this function assumes round_to is some power of 2.
    return (value + (round_to - 1)) & ~(round_to - 1)

def VCOS_ALIGN_DOWN(value, round_to):
    # Note: this function assumes round_to is some power of 2.
    return value & ~(round_to - 1)

# vc_image_types.h ###########################################################

class VC_RECT_T(ct.Structure):
    _fields_ = [
        ('x', ct.c_int32),
        ('y', ct.c_int32),
        ('width', ct.c_int32),
        ('height', ct.c_int32),
        ]

VC_IMAGE_TYPE_T = ct.c_uint32 # enum
(
   VC_IMAGE_MIN,
   VC_IMAGE_RGB565,
   VC_IMAGE_1BPP,
   VC_IMAGE_YUV420,
   VC_IMAGE_48BPP,
   VC_IMAGE_RGB888,
   VC_IMAGE_8BPP,
   VC_IMAGE_4BPP,
   VC_IMAGE_3D32,
   VC_IMAGE_3D32B,
   VC_IMAGE_3D32MAT,
   VC_IMAGE_RGB2X9,
   VC_IMAGE_RGB666,
   VC_IMAGE_PAL4_OBSOLETE,
   VC_IMAGE_PAL8_OBSOLETE,
   VC_IMAGE_RGBA32,
   VC_IMAGE_YUV422,
   VC_IMAGE_RGBA565,
   VC_IMAGE_RGBA16,
   VC_IMAGE_YUV_UV,
   VC_IMAGE_TF_RGBA32,
   VC_IMAGE_TF_RGBX32,
   VC_IMAGE_TF_FLOAT,
   VC_IMAGE_TF_RGBA16,
   VC_IMAGE_TF_RGBA5551,
   VC_IMAGE_TF_RGB565,
   VC_IMAGE_TF_YA88,
   VC_IMAGE_TF_BYTE,
   VC_IMAGE_TF_PAL8,
   VC_IMAGE_TF_PAL4,
   VC_IMAGE_TF_ETC1,
   VC_IMAGE_BGR888,
   VC_IMAGE_BGR888_NP,
   VC_IMAGE_BAYER,
   VC_IMAGE_CODEC,
   VC_IMAGE_YUV_UV32,
   VC_IMAGE_TF_Y8,
   VC_IMAGE_TF_A8,
   VC_IMAGE_TF_SHORT,
   VC_IMAGE_TF_1BPP,
   VC_IMAGE_OPENGL,
   VC_IMAGE_YUV444I,
   VC_IMAGE_YUV422PLANAR,
   VC_IMAGE_ARGB8888,
   VC_IMAGE_XRGB8888,
   VC_IMAGE_YUV422YUYV,
   VC_IMAGE_YUV422YVYU,
   VC_IMAGE_YUV422UYVY,
   VC_IMAGE_YUV422VYUY,
   VC_IMAGE_RGBX32,
   VC_IMAGE_RGBX8888,
   VC_IMAGE_BGRX8888,
   VC_IMAGE_YUV420SP,
   VC_IMAGE_YUV444PLANAR,
   VC_IMAGE_TF_U8,
   VC_IMAGE_TF_V8,
   VC_IMAGE_MAX,
) = range(57)

TRANSFORM_HFLIP = 1 << 0
TRANSFORM_VFLIP = 1 << 1
TRANSFORM_TRANSPOSE = 1 << 2

VC_IMAGE_TRANSFORM_T = ct.c_uint32 # enum
VC_IMAGE_ROT0          = 0
VC_IMAGE_MIRROR_ROT0   = TRANSFORM_HFLIP
VC_IMAGE_MIRROR_ROT180 = TRANSFORM_VFLIP
VC_IMAGE_ROT180        = TRANSFORM_HFLIP | TRANSFORM_VFLIP
VC_IMAGE_MIRROR_ROT90  = TRANSFORM_TRANSPOSE
VC_IMAGE_ROT270        = TRANSFORM_TRANSPOSE | TRANSFORM_HFLIP
VC_IMAGE_ROT90         = TRANSFORM_TRANSPOSE | TRANSFORM_VFLIP
VC_IMAGE_MIRROR_ROT270 = TRANSFORM_TRANSPOSE | TRANSFORM_HFLIP | TRANSFORM_VFLIP

VC_IMAGE_BAYER_ORDER_T = ct.c_uint32 # enum
(
   VC_IMAGE_BAYER_RGGB,
   VC_IMAGE_BAYER_GBRG,
   VC_IMAGE_BAYER_BGGR,
   VC_IMAGE_BAYER_GRBG,
) = range(4)

VC_IMAGE_BAYER_FORMAT_T = ct.c_uint32 # enum
(
   VC_IMAGE_BAYER_RAW6,
   VC_IMAGE_BAYER_RAW7,
   VC_IMAGE_BAYER_RAW8,
   VC_IMAGE_BAYER_RAW10,
   VC_IMAGE_BAYER_RAW12,
   VC_IMAGE_BAYER_RAW14,
   VC_IMAGE_BAYER_RAW16,
   VC_IMAGE_BAYER_RAW10_8,
   VC_IMAGE_BAYER_RAW12_8,
   VC_IMAGE_BAYER_RAW14_8,
   VC_IMAGE_BAYER_RAW10L,
   VC_IMAGE_BAYER_RAW12L,
   VC_IMAGE_BAYER_RAW14L,
   VC_IMAGE_BAYER_RAW16_BIG_ENDIAN,
   VC_IMAGE_BAYER_RAW4,
) = range(15)

# vc_display_types.h #########################################################

VCOS_DISPLAY_INPUT_FORMAT_T = ct.c_uint32 # enum
(
   VCOS_DISPLAY_INPUT_FORMAT_INVALID,
   VCOS_DISPLAY_INPUT_FORMAT_RGB888,
   VCOS_DISPLAY_INPUT_FORMAT_RGB565
) = range(3)

DISPLAY_INPUT_FORMAT_INVALID = VCOS_DISPLAY_INPUT_FORMAT_INVALID
DISPLAY_INPUT_FORMAT_RGB888  = VCOS_DISPLAY_INPUT_FORMAT_RGB888
DISPLAY_INPUT_FORMAT_RGB565  = VCOS_DISPLAY_INPUT_FORMAT_RGB565
DISPLAY_INPUT_FORMAT_T = VCOS_DISPLAY_INPUT_FORMAT_T

DISPLAY_3D_FORMAT_T = ct.c_uint32 # enum
(
   DISPLAY_3D_UNSUPPORTED,
   DISPLAY_3D_INTERLEAVED,
   DISPLAY_3D_SBS_FULL_AUTO,
   DISPLAY_3D_SBS_HALF_HORIZ,
   DISPLAY_3D_TB_HALF,
   DISPLAY_3D_FRAME_PACKING,
   DISPLAY_3D_FRAME_SEQUENTIAL,
   DISPLAY_3D_FORMAT_MAX,
) = range(8)

DISPLAY_INTERFACE_T = ct.c_uint32 # enum
(
   DISPLAY_INTERFACE_MIN,
   DISPLAY_INTERFACE_SMI,
   DISPLAY_INTERFACE_DPI,
   DISPLAY_INTERFACE_DSI,
   DISPLAY_INTERFACE_LVDS,
   DISPLAY_INTERFACE_MAX,
) = range(6)

DISPLAY_DITHER_T = ct.c_uint32 # enum
(
   DISPLAY_DITHER_NONE,
   DISPLAY_DITHER_RGB666,
   DISPLAY_DITHER_RGB565,
   DISPLAY_DITHER_RGB555,
   DISPLAY_DITHER_MAX,
) = range(5)

class DISPLAY_INFO_T(ct.Structure):
    _fields_ = [
        ('type',             DISPLAY_INTERFACE_T),
        ('width',            ct.c_uint32),
        ('height',           ct.c_uint32),
        ('input_format',     DISPLAY_INPUT_FORMAT_T),
        ('interlaced',       ct.c_uint32),
        ('output_dither',    DISPLAY_DITHER_T),
        ('pixel_freq',       ct.c_uint32),
        ('line_rate',        ct.c_uint32),
        ('format_3d',        DISPLAY_3D_FORMAT_T),
        ('use_pixelvalve_1', ct.c_uint32),
        ('dsi_video_mode',   ct.c_uint32),
        ('hvs_channel',      ct.c_uint32),
        ]

# vc_dispmanx_types.h ########################################################

DISPMANX_DISPLAY_HANDLE_T = ct.c_uint32
DISPMANX_UPDATE_HANDLE_T = ct.c_uint32
DISPMANX_ELEMENT_HANDLE_T = ct.c_uint32
DISPMANX_RESOURCE_HANDLE_T = ct.c_uint32
DISPMANX_PROTECTION_T = ct.c_uint32

DISPMANX_TRANSFORM_T = ct.c_uint32 # enum
DISPMANX_NO_ROTATE              = 0
DISPMANX_ROTATE_90              = 1
DISPMANX_ROTATE_180             = 2
DISPMANX_ROTATE_270             = 3
DISPMANX_FLIP_HRIZ              = 1 << 16
DISPMANX_FLIP_VERT              = 1 << 17
DISPMANX_STEREOSCOPIC_INVERT    = 1 << 19
DISPMANX_STEREOSCOPIC_NONE      = 0 << 20
DISPMANX_STEREOSCOPIC_MONO      = 1 << 20
DISPMANX_STEREOSCOPIC_SBS       = 2 << 20
DISPMANX_STEREOSCOPIC_TB        = 3 << 20
DISPMANX_STEREOSCOPIC_MASK      = 15 << 20
DISPMANX_SNAPSHOT_NO_YUV        = 1 << 24
DISPMANX_SNAPSHOT_NO_RGB        = 1 << 25
DISPMANX_SNAPSHOT_FILL          = 1 << 26
DISPMANX_SNAPSHOT_SWAP_RED_BLUE = 1 << 27
DISPMANX_SNAPSHOT_PACK          = 1 << 28

DISPMANX_FLAGS_ALPHA_T = ct.c_uint32 # enum
DISPMANX_FLAGS_ALPHA_FROM_SOURCE       = 0
DISPMANX_FLAGS_ALPHA_FIXED_ALL_PIXELS  = 1
DISPMANX_FLAGS_ALPHA_FIXED_NON_ZERO    = 2
DISPMANX_FLAGS_ALPHA_FIXED_EXCEED_0X07 = 3
DISPMANX_FLAGS_ALPHA_PREMULT           = 1 << 16
DISPMANX_FLAGS_ALPHA_MIX               = 1 << 17

class DISPMANX_ALPHA_T(ct.Structure):
    _fields_ = [
        ('flags',   DISPMANX_FLAGS_ALPHA_T),
        ('opacity', ct.c_uint32),
        ('mask',    ct.c_void_p),
        ]

class VC_DISPMANX_ALPHA_T(ct.Structure):
    _fields_ = [
        ('flags',   DISPMANX_FLAGS_ALPHA_T),
        ('opacity', ct.c_uint32),
        ('mask',    DISPMANX_RESOURCE_HANDLE_T),
        ]

DISPMANX_FLAGS_CLAMP_T = ct.c_uint32 # enum
(
  DISPMANX_FLAGS_CLAMP_NONE,
  DISPMANX_FLAGS_CLAMP_LUMA_TRANSPARENT,
  DISPMANX_FLAGS_CLAMP_TRANSPARENT,
  DISPMANX_FLAGS_CLAMP_REPLACE,
) = range(4)

DISPMANX_FLAGS_KEYMASK_T = ct.c_uint32 # enum
DISPMANX_FLAGS_KEYMASK_OVERRIDE = 1
DISPMANX_FLAGS_KEYMASK_SMOOTH   = 1 << 1
DISPMANX_FLAGS_KEYMASK_CR_INV   = 1 << 2
DISPMANX_FLAGS_KEYMASK_CB_INV   = 1 << 3
DISPMANX_FLAGS_KEYMASK_YY_INV   = 1 << 4

class _YUV(ct.Structure):
    _fields_ = [
        ('yy_upper', ct.c_uint8),
        ('yy_lower', ct.c_uint8),
        ('cr_upper', ct.c_uint8),
        ('cr_lower', ct.c_uint8),
        ('cb_upper', ct.c_uint8),
        ('cb_lower', ct.c_uint8),
        ]

class _RGB(ct.Structure):
    _fields_ = [
        ('red_upper',   ct.c_uint8),
        ('red_lower',   ct.c_uint8),
        ('green_upper', ct.c_uint8),
        ('green_lower', ct.c_uint8),
        ('blue_upper',  ct.c_uint8),
        ('blue_lower',  ct.c_uint8),
        ]

class DISPMANX_CLAMP_KEYS_T(ct.Union):
    _fields_ = [
        ('yuv', _YUV),
        ('rgb', _RGB),
        ]

class DISPMANX_CLAMP_T(ct.Structure):
    _fields_ = [
        ('mode',          DISPMANX_FLAGS_CLAMP_T),
        ('key_mask',      DISPMANX_FLAGS_KEYMASK_T),
        ('key_value',     DISPMANX_CLAMP_KEYS_T),
        ('replace_value', ct.c_uint32),
        ]

class DISPMANX_MODEINFO_T(ct.Structure):
    _fields_ = [
        ('width',        ct.c_int32),
        ('height',       ct.c_int32),
        ('transform',    DISPMANX_TRANSFORM_T),
        ('input_format', DISPLAY_INPUT_FORMAT_T),
        ('display_num',  ct.c_uint32),
        ]

DISPMANX_CALLBACK_FUNC_T = ct.CFUNCTYPE(
    None,
    DISPMANX_UPDATE_HANDLE_T, ct.c_void_p)

DISPMANX_PROGRESS_CALLBACK_FUNC_T = ct.CFUNCTYPE(
    None,
    DISPMANX_UPDATE_HANDLE_T, ct.c_uint32, ct.c_void_p)

# vc_dispmanx.h ##############################################################

vc_dispmanx_stop = _lib.vc_dispmanx_stop
vc_dispmanx_stop.argtypes = []
vc_dispmanx_stop.restype = None

vc_dispmanx_rect_set = _lib.vc_dispmanx_rect_set
vc_dispmanx_rect_set.argtypes = [ct.POINTER(VC_RECT_T), ct.c_uint32, ct.c_uint32, ct.c_uint32, ct.c_uint32]
vc_dispmanx_rect_set.restype = ct.c_int

vc_dispmanx_resource_create = _lib.vc_dispmanx_resource_create
vc_dispmanx_resource_create.argtypes = [VC_IMAGE_TYPE_T, ct.c_uint32, ct.c_uint32, ct.POINTER(ct.c_uint32)]
vc_dispmanx_resource_create.restype = DISPMANX_RESOURCE_HANDLE_T

vc_dispmanx_resource_write_data = _lib.vc_dispmanx_resource_write_data
vc_dispmanx_resource_write_data.argtypes = [DISPMANX_RESOURCE_HANDLE_T, VC_IMAGE_TYPE_T, ct.c_int, ct.c_void_p, ct.POINTER(VC_RECT_T)]
vc_dispmanx_resource_write_data.restype = ct.c_int

vc_dispmanx_resource_read_data = _lib.vc_dispmanx_resource_read_data
vc_dispmanx_resource_read_data.argtypes = [DISPMANX_RESOURCE_HANDLE_T, ct.POINTER(VC_RECT_T), ct.c_void_p, ct.c_uint32]
vc_dispmanx_resource_read_data.restype = ct.c_int

vc_dispmanx_resource_delete = _lib.vc_dispmanx_resource_delete
vc_dispmanx_resource_delete.argtypes = [DISPMANX_RESOURCE_HANDLE_T]
vc_dispmanx_resource_delete.restype = ct.c_int

vc_dispmanx_display_open = _lib.vc_dispmanx_display_open
vc_dispmanx_display_open.argtypes = [ct.c_uint32]
vc_dispmanx_display_open.restype = DISPMANX_DISPLAY_HANDLE_T

vc_dispmanx_display_open_mode = _lib.vc_dispmanx_display_open_mode
vc_dispmanx_display_open_mode.argtypes = [ct.c_uint32, ct.c_uint32]
vc_dispmanx_display_open_mode.restype = DISPMANX_DISPLAY_HANDLE_T

vc_dispmanx_display_open_offscreen = _lib.vc_dispmanx_display_open_offscreen
vc_dispmanx_display_open_offscreen.argtypes = [DISPMANX_RESOURCE_HANDLE_T, DISPMANX_TRANSFORM_T]
vc_dispmanx_display_open_offscreen.restype = DISPMANX_DISPLAY_HANDLE_T

vc_dispmanx_display_reconfigure = _lib.vc_dispmanx_display_reconfigure
vc_dispmanx_display_reconfigure.argtypes = [DISPMANX_DISPLAY_HANDLE_T, ct.c_uint32]
vc_dispmanx_display_reconfigure.restype = ct.c_int

vc_dispmanx_display_set_destination = _lib.vc_dispmanx_display_set_destination
vc_dispmanx_display_set_destination.argtypes = [DISPMANX_DISPLAY_HANDLE_T, DISPMANX_RESOURCE_HANDLE_T]
vc_dispmanx_display_set_destination.restype = ct.c_int

vc_dispmanx_display_set_background = _lib.vc_dispmanx_display_set_background
vc_dispmanx_display_set_background.argtypes = [DISPMANX_UPDATE_HANDLE_T, DISPMANX_DISPLAY_HANDLE_T, ct.c_uint8, ct.c_uint8, ct.c_uint8]
vc_dispmanx_display_set_background.restype = ct.c_int

vc_dispmanx_display_get_info = _lib.vc_dispmanx_display_get_info
vc_dispmanx_display_get_info.argtypes = [DISPMANX_DISPLAY_HANDLE_T, ct.POINTER(DISPMANX_MODEINFO_T)]
vc_dispmanx_display_get_info.restype = ct.c_int

vc_dispmanx_display_close = _lib.vc_dispmanx_display_close
vc_dispmanx_display_close.argtypes = [DISPMANX_DISPLAY_HANDLE_T]
vc_dispmanx_display_close.restype = ct.c_int

vc_dispmanx_update_start = _lib.vc_dispmanx_update_start
vc_dispmanx_update_start.argtypes = [ct.c_int32]
vc_dispmanx_update_start.restype = DISPMANX_UPDATE_HANDLE_T

vc_dispmanx_element_add = _lib.vc_dispmanx_element_add
vc_dispmanx_element_add.argtypes = [DISPMANX_UPDATE_HANDLE_T, DISPMANX_DISPLAY_HANDLE_T, ct.c_int32, ct.POINTER(VC_RECT_T), DISPMANX_RESOURCE_HANDLE_T, ct.POINTER(VC_RECT_T), DISPMANX_PROTECTION_T, VC_DISPMANX_ALPHA_T, DISPMANX_CLAMP_T, DISPMANX_TRANSFORM_T]
vc_dispmanx_element_add.restype = DISPMANX_ELEMENT_HANDLE_T

vc_dispmanx_element_change_source = _lib.vc_dispmanx_element_change_source
vc_dispmanx_element_change_source.argtypes = [DISPMANX_UPDATE_HANDLE_T, DISPMANX_ELEMENT_HANDLE_T, DISPMANX_RESOURCE_HANDLE_T]
vc_dispmanx_element_change_source.restype = ct.c_int

vc_dispmanx_element_change_layer = _lib.vc_dispmanx_element_change_layer
vc_dispmanx_element_change_layer.argtypes = [DISPMANX_UPDATE_HANDLE_T, DISPMANX_ELEMENT_HANDLE_T, ct.c_int32]
vc_dispmanx_element_change_layer.restype = ct.c_int

vc_dispmanx_element_modified = _lib.vc_dispmanx_element_modified
vc_dispmanx_element_modified.argtypes = [DISPMANX_UPDATE_HANDLE_T, DISPMANX_ELEMENT_HANDLE_T, ct.POINTER(VC_RECT_T)]
vc_dispmanx_element_modified.restype = ct.c_int

vc_dispmanx_element_remove = _lib.vc_dispmanx_element_remove
vc_dispmanx_element_remove.argtypes = [DISPMANX_UPDATE_HANDLE_T, DISPMANX_ELEMENT_HANDLE_T]
vc_dispmanx_element_remove.restype = ct.c_int

vc_dispmanx_update_submit = _lib.vc_dispmanx_update_submit
vc_dispmanx_update_submit.argtypes = [DISPMANX_UPDATE_HANDLE_T, DISPMANX_CALLBACK_FUNC_T, ct.c_void_p]
vc_dispmanx_update_submit.restype = ct.c_int

vc_dispmanx_update_submit_sync = _lib.vc_dispmanx_update_submit_sync
vc_dispmanx_update_submit_sync.argtypes = [DISPMANX_UPDATE_HANDLE_T]
vc_dispmanx_update_submit_sync.restype = ct.c_int

vc_dispmanx_query_image_formats = _lib.vc_dispmanx_query_image_formats
vc_dispmanx_query_image_formats.argtypes = [ct.POINTER(ct.c_uint32)]
vc_dispmanx_query_image_formats.restype = ct.c_int

vc_dispmanx_element_change_attributes = _lib.vc_dispmanx_element_change_attributes
vc_dispmanx_element_change_attributes.argtypes = [DISPMANX_UPDATE_HANDLE_T, DISPMANX_ELEMENT_HANDLE_T, ct.c_uint32, ct.c_int32, ct.c_uint8, ct.POINTER(VC_RECT_T), ct.POINTER(VC_RECT_T), DISPMANX_RESOURCE_HANDLE_T, DISPMANX_TRANSFORM_T]
vc_dispmanx_element_change_attributes.restype = ct.c_int

vc_vchi_dispmanx_init = _lib.vc_vchi_dispmanx_init
vc_vchi_dispmanx_init.argtypes = [VCHI_INSTANCE_T, ct.POINTER(VCHI_CONNECTION_T), ct.c_uint32]
vc_vchi_dispmanx_init.restype = None

vc_dispmanx_snapshot = _lib.vc_dispmanx_snapshot
vc_dispmanx_snapshot.argtypes = [DISPMANX_DISPLAY_HANDLE_T, DISPMANX_RESOURCE_HANDLE_T, DISPMANX_TRANSFORM_T]
vc_dispmanx_snapshot.restype = ct.c_int

vc_dispmanx_resource_set_palette = _lib.vc_dispmanx_resource_set_palette
vc_dispmanx_resource_set_palette.argtypes = [DISPMANX_RESOURCE_HANDLE_T, ct.c_void_p, ct.c_int, ct.c_int]
vc_dispmanx_resource_set_palette.restype = ct.c_int

vc_dispmanx_vsync_callback = _lib.vc_dispmanx_vsync_callback
vc_dispmanx_vsync_callback.argtypes = [DISPMANX_DISPLAY_HANDLE_T, DISPMANX_CALLBACK_FUNC_T, ct.c_void_p]
vc_dispmanx_vsync_callback.restype = ct.c_int

# vc_cec.h ###################################################################

CEC_BROADCAST_ADDR  = 0xF
CEC_TV_ADDRESS      = 0
CEC_MAX_XMIT_LENGTH = 15
CEC_CLEAR_ADDR      = 0xFFFF

CEC_VERSION                    = 0x4
CEC_VENDOR_ID_BROADCOM         = 0x18C086
CEC_VENDOR_ID_ONKYO            = 0x0009B0
CEC_VENDOR_ID_PANASONIC_EUROPE = 0x000F12
CEC_VENDOR_ID                  = 0

CEC_BLOCKING = 1
CEC_NONBLOCKING = 0

CEC_AllDevices_T = ct.c_uint32 # enum
(
   CEC_AllDevices_eTV,
   CEC_AllDevices_eRec1,
   CEC_AllDevices_eRec2,
   CEC_AllDevices_eSTB1,
   CEC_AllDevices_eDVD1,
   CEC_AllDevices_eAudioSystem,
   CEC_AllDevices_eSTB2,
   CEC_AllDevices_eSTB3,
   CEC_AllDevices_eDVD2,
   CEC_AllDevices_eRec3,
   CEC_AllDevices_eSTB4,
   CEC_AllDevices_eDVD3,
   CEC_AllDevices_eRsvd3,
   CEC_AllDevices_eRsvd4,
   CEC_AllDevices_eFreeUse,
   CEC_AllDevices_eUnRegistered,
) = range(16)

CEC_DEVICE_TYPE_T = ct.c_uint32 # enum
(
   CEC_DeviceType_TV,
   CEC_DeviceType_Rec,
   CEC_DeviceType_Reserved,
   CEC_DeviceType_Tuner,
   CEC_DeviceType_Playback,
   CEC_DeviceType_Audio,
   CEC_DeviceType_Switch,
   CEC_DeviceType_VidProc,
) = range(8)
CEC_DeviceType_Invalid = 0xF

CEC_OPCODE_T = ct.c_uint32 # enum
CEC_Opcode_FeatureAbort                = 0x00
CEC_Opcode_ImageViewOn                 = 0x04
CEC_Opcode_TunerStepIncrement          = 0x05
CEC_Opcode_TunerStepDecrement          = 0x06
CEC_Opcode_TunerDeviceStatus           = 0x07
CEC_Opcode_GiveTunerDeviceStatus       = 0x08
CEC_Opcode_RecordOn                    = 0x09
CEC_Opcode_RecordStatus                = 0x0A
CEC_Opcode_RecordOff                   = 0x0B
CEC_Opcode_TextViewOn                  = 0x0D
CEC_Opcode_RecordTVScreen              = 0x0F
CEC_Opcode_GiveDeckStatus              = 0x1A
CEC_Opcode_DeckStatus                  = 0x1B
CEC_Opcode_SetMenuLanguage             = 0x32
CEC_Opcode_ClearAnalogTimer            = 0x33
CEC_Opcode_SetAnalogTimer              = 0x34
CEC_Opcode_TimerStatus                 = 0x35
CEC_Opcode_Standby                     = 0x36
CEC_Opcode_Play                        = 0x41
CEC_Opcode_DeckControl                 = 0x42
CEC_Opcode_TimerClearedStatus          = 0x43
CEC_Opcode_UserControlPressed          = 0x44
CEC_Opcode_UserControlReleased         = 0x45
CEC_Opcode_GiveOSDName                 = 0x46
CEC_Opcode_SetOSDName                  = 0x47
CEC_Opcode_SetOSDString                = 0x64
CEC_Opcode_SetTimerProgramTitle        = 0x67
CEC_Opcode_SystemAudioModeRequest      = 0x70
CEC_Opcode_GiveAudioStatus             = 0x71
CEC_Opcode_SetSystemAudioMode          = 0x72
CEC_Opcode_ReportAudioStatus           = 0x7A
CEC_Opcode_GiveSystemAudioModeStatus   = 0x7D
CEC_Opcode_SystemAudioModeStatus       = 0x7E
CEC_Opcode_RoutingChange               = 0x80
CEC_Opcode_RoutingInformation          = 0x81
CEC_Opcode_ActiveSource                = 0x82
CEC_Opcode_GivePhysicalAddress         = 0x83
CEC_Opcode_ReportPhysicalAddress       = 0x84
CEC_Opcode_RequestActiveSource         = 0x85
CEC_Opcode_SetStreamPath               = 0x86
CEC_Opcode_DeviceVendorID              = 0x87
CEC_Opcode_VendorCommand               = 0x89
CEC_Opcode_VendorRemoteButtonDown      = 0x8A
CEC_Opcode_VendorRemoteButtonUp        = 0x8B
CEC_Opcode_GiveDeviceVendorID          = 0x8C
CEC_Opcode_MenuRequest                 = 0x8D
CEC_Opcode_MenuStatus                  = 0x8E
CEC_Opcode_GiveDevicePowerStatus       = 0x8F
CEC_Opcode_ReportPowerStatus           = 0x90
CEC_Opcode_GetMenuLanguage             = 0x91
CEC_Opcode_SelectAnalogService         = 0x92
CEC_Opcode_SelectDigitalService        = 0x93
CEC_Opcode_SetDigitalTimer             = 0x97
CEC_Opcode_ClearDigitalTimer           = 0x99
CEC_Opcode_SetAudioRate                = 0x9A
CEC_Opcode_InactiveSource              = 0x9D
CEC_Opcode_CECVersion                  = 0x9E
CEC_Opcode_GetCECVersion               = 0x9F
CEC_Opcode_VendorCommandWithID         = 0xA0
CEC_Opcode_ClearExternalTimer          = 0xA1
CEC_Opcode_SetExternalTimer            = 0xA2
CEC_Opcode_ReportShortAudioDescriptor  = 0xA3
CEC_Opcode_RequestShortAudioDescriptor = 0xA4
CEC_Opcode_InitARC                     = 0xC0
CEC_Opcode_ReportARCInited             = 0xC1
CEC_Opcode_ReportARCTerminated         = 0xC2
CEC_Opcode_RequestARCInit              = 0xC3
CEC_Opcode_RequestARCTermination       = 0xC4
CEC_Opcode_TerminateARC                = 0xC5
CEC_Opcode_CDC                         = 0xF8
CEC_Opcode_Abort                       = 0xFF

CEC_ABORT_REASON_T = ct.c_uint32 # enum
(
   CEC_Abort_Reason_Unrecognised_Opcode,
   CEC_Abort_Reason_Wrong_Mode,
   CEC_Abort_Reason_Cannot_Provide_Source,
   CEC_Abort_Reason_Invalid_Operand,
   CEC_Abort_Reason_Refused,
   CEC_Abort_Reason_Undetermined,
) = range(6)

CEC_DISPLAY_CONTROL_T = ct.c_uint32 # enum
CEC_DISPLAY_CONTROL_DEFAULT_TIME       = 0
CEC_DISPLAY_CONTROL_UNTIL_CLEARED      = 1 << 6
CEC_DISPLAY_CONTROL_CLEAR_PREV_MSG     = 1 << 7

CEC_POWER_STATUS_T = ct.c_uint32 # enum
(
   CEC_POWER_STATUS_ON,
   CEC_POWER_STATUS_STANDBY,
   CEC_POWER_STATUS_ON_PENDING,
   CEC_POWER_STATUS_STANDBY_PENDING,
) = range(4)

CEC_MENU_STATE_T = ct.c_uint32 # enum
(
   CEC_MENU_STATE_ACTIVATED,
   CEC_MENU_STATE_DEACTIVATED,
   CEC_MENU_STATE_QUERY,
) = range(3)

CEC_DECK_INFO_T = ct.c_uint32 # enum
(
   CEC_DECK_INFO_PLAY,
   CEC_DECK_INFO_RECORD,
   CEC_DECK_INFO_PLAY_REVERSE,
   CEC_DECK_INFO_STILL,
   CEC_DECK_INFO_SLOW,
   CEC_DECK_INFO_SLOW_REVERSE,
   CEC_DECK_INFO_SEARCH_FORWARD,
   CEC_DECK_INFO_SEARCH_REVERSE,
   CEC_DECK_INFO_NO_MEDIA,
   CEC_DECK_INFO_STOP,
   CEC_DECK_INFO_WIND,
   CEC_DECK_INFO_REWIND,
   CEC_DECK_IDX_SEARCH_FORWARD,
   CEC_DECK_IDX_SEARCH_REVERSE,
   CEC_DECK_OTHER_STATUS,
) = range(0x11, 0x20)

CEC_DECK_CTRL_MODE_T = ct.c_uint32 # enum
(
   CEC_DECK_CTRL_FORWARD,
   CEC_DECK_CTRL_BACKWARD,
   CEC_DECK_CTRL_STOP,
   CEC_DECK_CTRL_EJECT,
) = range(1, 5)

CEC_PLAY_MODE_T = ct.c_uint32 # enum
CEC_PLAY_FORWARD                       = 0x24
CEC_PLAY_REVERSE                       = 0x20
CEC_PLAY_STILL                         = 0x25
CEC_PLAY_SCAN_FORWARD_MIN_SPEED        = 0x05
CEC_PLAY_SCAN_FORWARD_MED_SPEED        = 0x06
CEC_PLAY_SCAN_FORWARD_MAX_SPEED        = 0x07
CEC_PLAY_SCAN_REVERSE_MIN_SPEED        = 0x09
CEC_PLAY_SCAN_REVERSE_MED_SPEED        = 0x0A
CEC_PLAY_SCAN_REVERSE_MAX_SPEED        = 0x0B
CEC_PLAY_SLOW_FORWARD_MIN_SPEED        = 0x15
CEC_PLAY_SLOW_FORWARD_MED_SPEED        = 0x16
CEC_PLAY_SLOW_FORWARD_MAX_SPEED        = 0x17
CEC_PLAY_SLOW_REVERSE_MIN_SPEED        = 0x19
CEC_PLAY_SLOW_REVERSE_MED_SPEED        = 0x1A
CEC_PLAY_SLOW_REVERSE_MAX_SPEED        = 0x1B

CEC_DECK_STATUS_REQUEST_T = ct.c_uint32 # enum
(
   CEC_DECK_STATUS_ON,
   CEC_DECK_STATUS_OFF,
   CEC_DECK_STATUS_ONCE,
) = range(1, 4)

CEC_USER_CONTROL_T = ct.c_uint32 # enum
CEC_User_Control_Select                      = 0x00
CEC_User_Control_Up                          = 0x01
CEC_User_Control_Down                        = 0x02
CEC_User_Control_Left                        = 0x03
CEC_User_Control_Right                       = 0x04
CEC_User_Control_RightUp                     = 0x05
CEC_User_Control_RightDown                   = 0x06
CEC_User_Control_LeftUp                      = 0x07
CEC_User_Control_LeftDown                    = 0x08
CEC_User_Control_RootMenu                    = 0x09
CEC_User_Control_SetupMenu                   = 0x0A
CEC_User_Control_ContentsMenu                = 0x0B
CEC_User_Control_FavoriteMenu                = 0x0C
CEC_User_Control_Exit                        = 0x0D
CEC_User_Control_Number0                     = 0x20
CEC_User_Control_Number1                     = 0x21
CEC_User_Control_Number2                     = 0x22
CEC_User_Control_Number3                     = 0x23
CEC_User_Control_Number4                     = 0x24
CEC_User_Control_Number5                     = 0x25
CEC_User_Control_Number6                     = 0x26
CEC_User_Control_Number7                     = 0x27
CEC_User_Control_Number8                     = 0x28
CEC_User_Control_Number9                     = 0x29
CEC_User_Control_Dot                         = 0x2A
CEC_User_Control_Enter                       = 0x2B
CEC_User_Control_Clear                       = 0x2C
CEC_User_Control_ChannelUp                   = 0x30
CEC_User_Control_ChannelDown                 = 0x31
CEC_User_Control_PreviousChannel             = 0x32
CEC_User_Control_SoundSelect                 = 0x33
CEC_User_Control_InputSelect                 = 0x34
CEC_User_Control_DisplayInformation          = 0x35
CEC_User_Control_Help                        = 0x36
CEC_User_Control_PageUp                      = 0x37
CEC_User_Control_PageDown                    = 0x38
CEC_User_Control_Power                       = 0x40
CEC_User_Control_VolumeUp                    = 0x41
CEC_User_Control_VolumeDown                  = 0x42
CEC_User_Control_Mute                        = 0x43
CEC_User_Control_Play                        = 0x44
CEC_User_Control_Stop                        = 0x45
CEC_User_Control_Pause                       = 0x46
CEC_User_Control_Record                      = 0x47
CEC_User_Control_Rewind                      = 0x48
CEC_User_Control_FastForward                 = 0x49
CEC_User_Control_Eject                       = 0x4A
CEC_User_Control_Forward                     = 0x4B
CEC_User_Control_Backward                    = 0x4C
CEC_User_Control_Angle                       = 0x50
CEC_User_Control_Subpicture                  = 0x51
CEC_User_Control_VideoOnDemand               = 0x52
CEC_User_Control_EPG                         = 0x53
CEC_User_Control_TimerProgramming            = 0x54
CEC_User_Control_InitialConfig               = 0x55
CEC_User_Control_PlayFunction                = 0x60
CEC_User_Control_PausePlayFunction           = 0x61
CEC_User_Control_RecordFunction              = 0x62
CEC_User_Control_PauseRecordFunction         = 0x63
CEC_User_Control_StopFunction                = 0x64
CEC_User_Control_MuteFunction                = 0x65
CEC_User_Control_RestoreVolumeFunction       = 0x66
CEC_User_Control_TuneFunction                = 0x67
CEC_User_Control_SelectDiskFunction          = 0x68
CEC_User_Control_SelectAVInputFunction       = 0x69
CEC_User_Control_SelectAudioInputFunction    = 0x6A
CEC_User_Control_F1Blue                      = 0x71
CEC_User_Control_F2Red                       = 0x72
CEC_User_Control_F3Green                     = 0x73
CEC_User_Control_F4Yellow                    = 0x74
CEC_User_Control_F5                          = 0x75

class VC_CEC_TOPOLOGY_T(ct.Structure):
    _fields_ = [
        ('active_mask', ct.c_uint16),
        ('num_devices', ct.c_uint16),
        ('device_attr', ct.c_uint32 * 16),
        ]

class VC_CEC_MESSAGE_T(ct.Structure):
    _fields_ = [
        ('length', ct.c_uint32),
        ('initiator', CEC_AllDevices_T),
        ('follower',  CEC_AllDevices_T),
        ('payload',   ct.c_uint8 * (CEC_MAX_XMIT_LENGTH + 1)),
        ]

VC_CEC_NOTIFY_T = ct.c_uint32 # enum
VC_CEC_NOTIFY_NONE       = 0
VC_CEC_TX                = 1 << 0
VC_CEC_RX                = 1 << 1
VC_CEC_BUTTON_PRESSED    = 1 << 2
VC_CEC_BUTTON_RELEASE    = 1 << 3
VC_CEC_REMOTE_PRESSED    = 1 << 4
VC_CEC_REMOTE_RELEASE    = 1 << 5
VC_CEC_LOGICAL_ADDR      = 1 << 6
VC_CEC_TOPOLOGY          = 1 << 7
VC_CEC_LOGICAL_ADDR_LOST = 1 << 15

CEC_CALLBACK_T = ct.CFUNCTYPE(
    None,
    ct.c_void_p, ct.c_uint32, ct.c_uint32, ct.c_uint32, ct.c_uint32)

CEC_CB_REASON     = lambda x: x & 0xFFFF
CEC_CB_MSG_LENGTH = lambda x: (x >> 16) & 0xFF
CEC_CB_RC         = lambda x: (x >> 24) & 0xFF

CEC_CB_INITIATOR  = lambda x: (x >> 4) & 0xF
CEC_CB_FOLLOWER   = lambda x: x & 0xF
CEC_CB_OPCODE     = lambda x: (x >> 8) & 0xFF
CEC_CB_OPERAND1   = lambda x: (x >> 16) & 0xFF
CEC_CB_OPERAND2   = lambda x: (x >> 24) & 0xFF

VC_CEC_ERROR_T = ct.c_uint32 # enum
(
   VC_CEC_SUCCESS,
   VC_CEC_ERROR_NO_ACK,
   VC_CEC_ERROR_SHUTDOWN,
   VC_CEC_ERROR_BUSY,
   VC_CEC_ERROR_NO_LA,
   VC_CEC_ERROR_NO_PA,
   VC_CEC_ERROR_NO_TOPO,
   VC_CEC_ERROR_INVALID_FOLLOWER,
   VC_CEC_ERROR_INVALID_ARGUMENT,
) = range(9)

# vc_cecservice.h ############################################################

CECSERVICE_CALLBACK_T = ct.CFUNCTYPE(
    None,
    ct.c_void_p, ct.c_uint32, ct.c_uint32, ct.c_uint32, ct.c_uint32, ct.c_uint32)

vc_vchi_cec_init = _lib.vc_vchi_cec_init
vc_vchi_cec_init.argtypes = [VCHI_INSTANCE_T, ct.POINTER(ct.POINTER(VCHI_CONNECTION_T)), ct.c_uint32]
vc_vchi_cec_init.restype = None

vc_vchi_cec_stop = _lib.vc_vchi_cec_stop
vc_vchi_cec_stop.argtypes = []
vc_vchi_cec_stop.restype = None

vc_cec_register_callback = _lib.vc_cec_register_callback
vc_cec_register_callback.argtypes = [CECSERVICE_CALLBACK_T, ct.c_void_p]
vc_cec_register_callback.restype = None

vc_cec_register_command = _lib.vc_cec_register_command
vc_cec_register_command.argtypes = [CEC_OPCODE_T]
vc_cec_register_command.restype = ct.c_int

vc_cec_register_all = _lib.vc_cec_register_all
vc_cec_register_all.argtypes = []
vc_cec_register_all.restype = ct.c_int

vc_cec_deregister_command = _lib.vc_cec_deregister_command
vc_cec_deregister_command.argtypes = [CEC_OPCODE_T]
vc_cec_deregister_command.restype = ct.c_int

vc_cec_deregister_all = _lib.vc_cec_deregister_all
vc_cec_deregister_all.argtypes = []
vc_cec_deregister_all.restype = ct.c_int

vc_cec_send_message = _lib.vc_cec_send_message
vc_cec_send_message.argtypes = [ct.c_uint32, ct.POINTER(ct.c_uint8), ct.c_uint32, vcos_bool_t]
vc_cec_send_message.restype = ct.c_int

vc_cec_get_logical_address = _lib.vc_cec_get_logical_address
vc_cec_get_logical_address.argtypes = [ct.POINTER(CEC_AllDevices_T)]
vc_cec_get_logical_address.restype = ct.c_int

vc_cec_alloc_logical_address = _lib.vc_cec_alloc_logical_address
vc_cec_alloc_logical_address.argtypes = []
vc_cec_alloc_logical_address.restype = ct.c_int

vc_cec_release_logical_address = _lib.vc_cec_release_logical_address
vc_cec_release_logical_address.argtypes = []
vc_cec_release_logical_address.restype = ct.c_int

vc_cec_get_topology = _lib.vc_cec_get_topology
vc_cec_get_topology.argtypes = [ct.POINTER(VC_CEC_TOPOLOGY_T)]
vc_cec_get_topology.restype = ct.c_int

vc_cec_set_vendor_id = _lib.vc_cec_set_vendor_id
vc_cec_set_vendor_id.argtypes = [ct.c_uint32]
vc_cec_set_vendor_id.restype = ct.c_int

vc_cec_set_osd_name = _lib.vc_cec_set_osd_name
vc_cec_set_osd_name.argtypes = [ct.c_char_p]
vc_cec_set_osd_name.restype = ct.c_int

vc_cec_get_physical_address = _lib.vc_cec_get_physical_address
vc_cec_get_physical_address.argtypes = [ct.POINTER(ct.c_uint16)]
vc_cec_get_physical_address.restype = ct.c_int

vc_cec_get_vendor_id = _lib.vc_cec_get_vendor_id
vc_cec_get_vendor_id.argtypes = [CEC_AllDevices_T, ct.POINTER(ct.c_uint32)]
vc_cec_get_vendor_id.restype = ct.c_int

vc_cec_device_type = _lib.vc_cec_device_type
vc_cec_device_type.argtypes = [CEC_AllDevices_T]
vc_cec_device_type.restype = CEC_DEVICE_TYPE_T

vc_cec_send_message2 = _lib.vc_cec_send_message2
vc_cec_send_message2.argtypes = [ct.POINTER(VC_CEC_MESSAGE_T)]
vc_cec_send_message2.restype = ct.c_int

vc_cec_param2message = _lib.vc_cec_param2message
vc_cec_param2message.argtypes = [ct.c_uint32, ct.c_uint32, ct.c_uint32, ct.c_uint32, ct.c_uint32, ct.POINTER(VC_CEC_MESSAGE_T)]
vc_cec_param2message.restype = ct.c_int

vc_cec_poll_address = _lib.vc_cec_poll_address
vc_cec_poll_address.argtypes = [CEC_AllDevices_T]
vc_cec_poll_address.restype = ct.c_int

vc_cec_set_logical_address = _lib.vc_cec_set_logical_address
vc_cec_set_logical_address.argtypes = [CEC_AllDevices_T, CEC_DEVICE_TYPE_T, ct.c_uint32]
vc_cec_set_logical_address.restype = ct.c_int

vc_cec_add_device = _lib.vc_cec_add_device
vc_cec_add_device.argtypes = [CEC_AllDevices_T, ct.c_uint16, CEC_DEVICE_TYPE_T, vcos_bool_t]
vc_cec_add_device.restype = ct.c_int

vc_cec_set_passive = _lib.vc_cec_set_passive
vc_cec_set_passive.argtypes = [vcos_bool_t]
vc_cec_set_passive.restype = ct.c_int

vc_cec_send_FeatureAbort = _lib.vc_cec_send_FeatureAbort
vc_cec_send_FeatureAbort.argtypes = [ct.c_uint32, CEC_OPCODE_T, CEC_ABORT_REASON_T]
vc_cec_send_FeatureAbort.restype = ct.c_int

vc_cec_send_ActiveSource = _lib.vc_cec_send_ActiveSource
vc_cec_send_ActiveSource.argtypes = [ct.c_uint16, vcos_bool_t]
vc_cec_send_ActiveSource.restype = ct.c_int

vc_cec_send_ImageViewOn = _lib.vc_cec_send_ImageViewOn
vc_cec_send_ImageViewOn.argtypes = [ct.c_uint32, vcos_bool_t]
vc_cec_send_ImageViewOn.restype = ct.c_int

vc_cec_send_SetOSDString = _lib.vc_cec_send_SetOSDString
vc_cec_send_SetOSDString.argtypes = [ct.c_uint32, CEC_DISPLAY_CONTROL_T, ct.c_char_p, vcos_bool_t]
vc_cec_send_SetOSDString.restype = ct.c_int

vc_cec_send_Standby = _lib.vc_cec_send_Standby
vc_cec_send_Standby.argtypes = [ct.c_uint32, vcos_bool_t]
vc_cec_send_Standby.restype = ct.c_int

vc_cec_send_MenuStatus = _lib.vc_cec_send_MenuStatus
vc_cec_send_MenuStatus.argtypes = [ct.c_uint32, CEC_MENU_STATE_T, vcos_bool_t]
vc_cec_send_MenuStatus.restype = ct.c_int

vc_cec_send_ReportPhysicalAddress = _lib.vc_cec_send_ReportPhysicalAddress
vc_cec_send_ReportPhysicalAddress.argtypes = [ct.c_uint16, CEC_DEVICE_TYPE_T, vcos_bool_t]
vc_cec_send_ReportPhysicalAddress.restype = ct.c_int

# vc_gencmd.h ################################################################

vc_gencmd_init = _lib.vc_gencmd_init
vc_gencmd_init.argtypes = []
vc_gencmd_init.restype = ct.c_int

vc_gencmd_stop = _lib.vc_gencmd_stop
vc_gencmd_stop.argtypes = []
vc_gencmd_stop.restype = None

vc_gencmd_send = _lib.vc_gencmd_send
vc_gencmd_send.argtypes = [ct.c_char_p]
vc_gencmd_send.restype = ct.c_int

vc_gencmd_read_response = _lib.vc_gencmd_read_response
vc_gencmd_read_response.argtypes = [ct.c_char_p, ct.c_int]
vc_gencmd_read_response.restype = ct.c_int

vc_gencmd = _lib.vc_gencmd
vc_gencmd.argtypes = [ct.c_char_p, ct.c_int, ct.c_char_p]
vc_gencmd.restype = ct.c_int

vc_gencmd_string_property = _lib.vc_gencmd_string_property
vc_gencmd_string_property.argtypes = [ct.c_char_p, ct.c_char_p, ct.POINTER(ct.c_char_p), ct.POINTER(ct.c_int)]
vc_gencmd_string_property.restype = ct.c_int

vc_gencmd_number_property = _lib.vc_gencmd_number_property
vc_gencmd_number_property.argtypes = [ct.c_char_p, ct.c_char_p, ct.POINTER(ct.c_int)]
vc_gencmd_number_property.restype = ct.c_int

vc_gencmd_until = _lib.vc_gencmd_until
vc_gencmd_until.argtypes = [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_int]
vc_gencmd_until.restype = ct.c_int




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Python camera library for the Rasperry-Pi camera module
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    print_function,
    division,
    absolute_import,
    )

# Make Py2's str and range equivalent to Py3's
native_str = str
str = type('')
try:
    range = xrange
except NameError:
    pass

import io
import ctypes as ct
import warnings

import numpy as np
from numpy.lib.stride_tricks import as_strided

from . import mmalobj as mo, mmal
from .exc import (
    mmal_check,
    PiCameraValueError,
    PiCameraDeprecated,
    PiCameraPortDisabled,
    )


motion_dtype = np.dtype([
    (native_str('x'),   np.int8),
    (native_str('y'),   np.int8),
    (native_str('sad'), np.uint16),
    ])


def raw_resolution(resolution, splitter=False):
    """
    Round a (width, height) tuple up to the nearest multiple of 32 horizontally
    and 16 vertically (as this is what the Pi's camera module does for
    unencoded output).
    """
    width, height = resolution
    if splitter:
        fwidth = (width + 15) & ~15
    else:
        fwidth = (width + 31) & ~31
    fheight = (height + 15) & ~15
    return fwidth, fheight


def bytes_to_yuv(data, resolution):
    """
    Converts a bytes object containing YUV data to a `numpy`_ array.
    """
    width, height = resolution
    fwidth, fheight = raw_resolution(resolution)
    y_len = fwidth * fheight
    uv_len = (fwidth // 2) * (fheight // 2)
    if len(data) != (y_len + 2 * uv_len):
        raise PiCameraValueError(
            'Incorrect buffer length for resolution %dx%d' % (width, height))
    # Separate out the Y, U, and V values from the array
    a = np.frombuffer(data, dtype=np.uint8)
    Y = a[:y_len].reshape((fheight, fwidth))
    Uq = a[y_len:-uv_len].reshape((fheight // 2, fwidth // 2))
    Vq = a[-uv_len:].reshape((fheight // 2, fwidth // 2))
    # Reshape the values into two dimensions, and double the size of the
    # U and V values (which only have quarter resolution in YUV4:2:0)
    U = np.empty_like(Y)
    V = np.empty_like(Y)
    U[0::2, 0::2] = Uq
    U[0::2, 1::2] = Uq
    U[1::2, 0::2] = Uq
    U[1::2, 1::2] = Uq
    V[0::2, 0::2] = Vq
    V[0::2, 1::2] = Vq
    V[1::2, 0::2] = Vq
    V[1::2, 1::2] = Vq
    # Stack the channels together and crop to the actual resolution
    return np.dstack((Y, U, V))[:height, :width]


def bytes_to_rgb(data, resolution):
    """
    Converts a bytes objects containing RGB/BGR data to a `numpy`_ array.
    """
    width, height = resolution
    fwidth, fheight = raw_resolution(resolution)
    # Workaround: output from the video splitter is rounded to 16x16 instead
    # of 32x16 (but only for RGB, and only when a resizer is not used)
    if len(data) != (fwidth * fheight * 3):
        fwidth, fheight = raw_resolution(resolution, splitter=True)
        if len(data) != (fwidth * fheight * 3):
            raise PiCameraValueError(
                'Incorrect buffer length for resolution %dx%d' % (width, height))
    # Crop to the actual resolution
    return np.frombuffer(data, dtype=np.uint8).\
            reshape((fheight, fwidth, 3))[:height, :width, :]


class PiArrayOutput(io.BytesIO):
    """
    Base class for capture arrays.

    This class extends :class:`io.BytesIO` with a `numpy`_ array which is
    intended to be filled when :meth:`~io.IOBase.flush` is called (i.e. at the
    end of capture).

    .. attribute:: array

        After :meth:`~io.IOBase.flush` is called, this attribute contains the
        frame's data as a multi-dimensional `numpy`_ array. This is typically
        organized with the dimensions ``(rows, columns, plane)``. Hence, an
        RGB image with dimensions *x* and *y* would produce an array with shape
        ``(y, x, 3)``.
    """

    def __init__(self, camera, size=None):
        super(PiArrayOutput, self).__init__()
        self.camera = camera
        self.size = size
        self.array = None

    def close(self):
        super(PiArrayOutput, self).close()
        self.array = None

    def truncate(self, size=None):
        """
        Resize the stream to the given size in bytes (or the current position
        if size is not specified). This resizing can extend or reduce the
        current file size.  The new file size is returned.

        In prior versions of picamera, truncation also changed the position of
        the stream (because prior versions of these stream classes were
        non-seekable). This functionality is now deprecated; scripts should
        use :meth:`~io.IOBase.seek` and :meth:`truncate` as one would with
        regular :class:`~io.BytesIO` instances.
        """
        if size is not None:
            warnings.warn(
                PiCameraDeprecated(
                    'This method changes the position of the stream to the '
                    'truncated length; this is deprecated functionality and '
                    'you should not rely on it (seek before or after truncate '
                    'to ensure position is consistent)'))
        super(PiArrayOutput, self).truncate(size)
        if size is not None:
            self.seek(size)


class PiRGBArray(PiArrayOutput):
    """
    Produces a 3-dimensional RGB array from an RGB capture.

    This custom output class can be used to easily obtain a 3-dimensional numpy
    array, organized (rows, columns, colors), from an unencoded RGB capture.
    The array is accessed via the :attr:`~PiArrayOutput.array` attribute. For
    example::

        import picamera
        import picamera.array

        with picamera.PiCamera() as camera:
            with picamera.array.PiRGBArray(camera) as output:
                camera.capture(output, 'rgb')
                print('Captured %dx%d image' % (
                        output.array.shape[1], output.array.shape[0]))

    You can re-use the output to produce multiple arrays by emptying it with
    ``truncate(0)`` between captures::

        import picamera
        import picamera.array

        with picamera.PiCamera() as camera:
            with picamera.array.PiRGBArray(camera) as output:
                camera.resolution = (1280, 720)
                camera.capture(output, 'rgb')
                print('Captured %dx%d image' % (
                        output.array.shape[1], output.array.shape[0]))
                output.truncate(0)
                camera.resolution = (640, 480)
                camera.capture(output, 'rgb')
                print('Captured %dx%d image' % (
                        output.array.shape[1], output.array.shape[0]))

    If you are using the GPU resizer when capturing (with the *resize*
    parameter of the various :meth:`~PiCamera.capture` methods), specify the
    resized resolution as the optional *size* parameter when constructing the
    array output::

        import picamera
        import picamera.array

        with picamera.PiCamera() as camera:
            camera.resolution = (1280, 720)
            with picamera.array.PiRGBArray(camera, size=(640, 360)) as output:
                camera.capture(output, 'rgb', resize=(640, 360))
                print('Captured %dx%d image' % (
                        output.array.shape[1], output.array.shape[0]))
    """

    def flush(self):
        super(PiRGBArray, self).flush()
        self.array = bytes_to_rgb(self.getvalue(), self.size or self.camera.resolution)


class PiYUVArray(PiArrayOutput):
    """
    Produces 3-dimensional YUV & RGB arrays from a YUV capture.

    This custom output class can be used to easily obtain a 3-dimensional numpy
    array, organized (rows, columns, channel), from an unencoded YUV capture.
    The array is accessed via the :attr:`~PiArrayOutput.array` attribute. For
    example::

        import picamera
        import picamera.array

        with picamera.PiCamera() as camera:
            with picamera.array.PiYUVArray(camera) as output:
                camera.capture(output, 'yuv')
                print('Captured %dx%d image' % (
                        output.array.shape[1], output.array.shape[0]))

    The :attr:`rgb_array` attribute can be queried for the equivalent RGB
    array (conversion is performed using the `ITU-R BT.601`_ matrix)::

        import picamera
        import picamera.array

        with picamera.PiCamera() as camera:
            with picamera.array.PiYUVArray(camera) as output:
                camera.resolution = (1280, 720)
                camera.capture(output, 'yuv')
                print(output.array.shape)
                print(output.rgb_array.shape)

    If you are using the GPU resizer when capturing (with the *resize*
    parameter of the various :meth:`~picamera.PiCamera.capture` methods),
    specify the resized resolution as the optional *size* parameter when
    constructing the array output::

        import picamera
        import picamera.array

        with picamera.PiCamera() as camera:
            camera.resolution = (1280, 720)
            with picamera.array.PiYUVArray(camera, size=(640, 360)) as output:
                camera.capture(output, 'yuv', resize=(640, 360))
                print('Captured %dx%d image' % (
                        output.array.shape[1], output.array.shape[0]))

    .. _ITU-R BT.601: https://en.wikipedia.org/wiki/YCbCr#ITU-R_BT.601_conversion
    """

    def __init__(self, camera, size=None):
        super(PiYUVArray, self).__init__(camera, size)
        self._rgb = None

    def flush(self):
        super(PiYUVArray, self).flush()
        self.array = bytes_to_yuv(self.getvalue(), self.size or self.camera.resolution)
        self._rgb = None

    @property
    def rgb_array(self):
        if self._rgb is None:
            # Apply the standard biases
            YUV = self.array.astype(float)
            YUV[:, :, 0]  = YUV[:, :, 0]  - 16  # Offset Y by 16
            YUV[:, :, 1:] = YUV[:, :, 1:] - 128 # Offset UV by 128
            # YUV conversion matrix from ITU-R BT.601 version (SDTV)
            #              Y       U       V
            M = np.array([[1.164,  0.000,  1.596],    # R
                          [1.164, -0.392, -0.813],    # G
                          [1.164,  2.017,  0.000]])   # B
            # Calculate the dot product with the matrix to produce RGB output,
            # clamp the results to byte range and convert to bytes
            self._rgb = YUV.dot(M.T).clip(0, 255).astype(np.uint8)
        return self._rgb


class BroadcomRawHeader(ct.Structure):
    _fields_ = [
        ('name',          ct.c_char * 32),
        ('width',         ct.c_uint16),
        ('height',        ct.c_uint16),
        ('padding_right', ct.c_uint16),
        ('padding_down',  ct.c_uint16),
        ('dummy',         ct.c_uint32 * 6),
        ('transform',     ct.c_uint16),
        ('format',        ct.c_uint16),
        ('bayer_order',   ct.c_uint8),
        ('bayer_format',  ct.c_uint8),
        ]


class PiBayerArray(PiArrayOutput):
    """
    Produces a 3-dimensional RGB array from raw Bayer data.

    This custom output class is intended to be used with the
    :meth:`~picamera.PiCamera.capture` method, with the *bayer* parameter set
    to ``True``, to include raw Bayer data in the JPEG output.  The class
    strips out the raw data, and constructs a numpy array from it.  The
    resulting data is accessed via the :attr:`~PiArrayOutput.array` attribute::

        import picamera
        import picamera.array

        with picamera.PiCamera() as camera:
            with picamera.array.PiBayerArray(camera) as output:
                camera.capture(output, 'jpeg', bayer=True)
                print(output.array.shape)

    The *output_dims* parameter specifies whether the resulting array is
    three-dimensional (the default, or when *output_dims* is 3), or
    two-dimensional (when *output_dims* is 2). The three-dimensional data is
    already separated into the three color planes, whilst the two-dimensional
    variant is not (in which case you need to know the Bayer ordering to
    accurately deal with the results).

    .. note::

        Bayer data is *usually* full resolution, so the resulting array usually
        has the shape (1944, 2592, 3) with the V1 module, or (2464, 3280, 3)
        with the V2 module (if two-dimensional output is requested the
        3-layered color dimension is omitted). If the camera's
        :attr:`~picamera.PiCamera.sensor_mode` has been forced to something
        other than 0, then the output will be the native size for the requested
        sensor mode.

        This also implies that the optional *size* parameter (for specifying a
        resizer resolution) is not available with this array class.

    As the sensor records 10-bit values, the array uses the unsigned 16-bit
    integer data type.

    By default, `de-mosaicing`_ is **not** performed; if the resulting array is
    viewed it will therefore appear dark and too green (due to the green bias
    in the `Bayer pattern`_). A trivial weighted-average demosaicing algorithm
    is provided in the :meth:`demosaic` method::

        import picamera
        import picamera.array

        with picamera.PiCamera() as camera:
            with picamera.array.PiBayerArray(camera) as output:
                camera.capture(output, 'jpeg', bayer=True)
                print(output.demosaic().shape)

    Viewing the result of the de-mosaiced data will look more normal but still
    considerably worse quality than the regular camera output (as none of the
    other usual post-processing steps like auto-exposure, white-balance,
    vignette compensation, and smoothing have been performed).

    .. versionchanged:: 1.13
        This class now supports the V2 module properly, and handles flipped
        images, and forced sensor modes correctly.

    .. _de-mosaicing: https://en.wikipedia.org/wiki/Demosaicing
    .. _Bayer pattern: https://en.wikipedia.org/wiki/Bayer_filter
    """
    BAYER_OFFSETS = {
        0: ((0, 0), (1, 0), (0, 1), (1, 1)),
        1: ((1, 0), (0, 0), (1, 1), (0, 1)),
        2: ((1, 1), (0, 1), (1, 0), (0, 0)),
        3: ((0, 1), (1, 1), (0, 0), (1, 0)),
        }

    def __init__(self, camera, output_dims=3):
        super(PiBayerArray, self).__init__(camera, size=None)
        if not (2 <= output_dims <= 3):
            raise PiCameraValueError('output_dims must be 2 or 3')
        self._demo = None
        self._header = None
        self._output_dims = output_dims

    @property
    def output_dims(self):
        return self._output_dims

    def _to_3d(self, array):
        array_3d = np.zeros(array.shape + (3,), dtype=array.dtype)
        (
            (ry, rx), (gy, gx), (Gy, Gx), (by, bx)
            ) = PiBayerArray.BAYER_OFFSETS[self._header.bayer_order]
        array_3d[ry::2, rx::2, 0] = array[ry::2, rx::2] # Red
        array_3d[gy::2, gx::2, 1] = array[gy::2, gx::2] # Green
        array_3d[Gy::2, Gx::2, 1] = array[Gy::2, Gx::2] # Green
        array_3d[by::2, bx::2, 2] = array[by::2, bx::2] # Blue
        return array_3d

    def flush(self):
        super(PiBayerArray, self).flush()
        self._demo = None
        offset = {
            'OV5647': {
                0: 6404096,
                1: 2717696,
                2: 6404096,
                3: 6404096,
                4: 1625600,
                5: 1233920,
                6: 445440,
                7: 445440,
                },
            'IMX219': {
                0: 10270208,
                1: 2678784,
                2: 10270208,
                3: 10270208,
                4: 2628608,
                5: 1963008,
                6: 1233920,
                7: 445440,
                },
            }[self.camera.revision.upper()][self.camera.sensor_mode]
        data = self.getvalue()[-offset:]
        if data[:4] != b'BRCM':
            raise PiCameraValueError('Unable to locate Bayer data at end of buffer')
        # Extract header (with bayer order and other interesting bits), which
        # is 176 bytes from start of bayer data, and pixel data which 32768
        # bytes from start of bayer data
        self._header = BroadcomRawHeader.from_buffer_copy(
            data[176:176 + ct.sizeof(BroadcomRawHeader)])
        data = np.frombuffer(data, dtype=np.uint8, offset=32768)
        # Reshape and crop the data. The crop's width is multiplied by 5/4 to
        # deal with the packed 10-bit format; the shape's width is calculated
        # in a similar fashion but with padding included (which involves
        # several additional padding steps)
        crop = mo.PiResolution(
            self._header.width * 5 // 4,
            self._header.height)
        shape = mo.PiResolution(
            (((self._header.width + self._header.padding_right) * 5) + 3) // 4,
            (self._header.height + self._header.padding_down)
            ).pad()
        data = data.reshape((shape.height, shape.width))[:crop.height, :crop.width]
        # Unpack 10-bit values; every 5 bytes contains the high 8-bits of 4
        # values followed by the low 2-bits of 4 values packed into the fifth
        # byte
        data = data.astype(np.uint16) << 2
        for byte in range(4):
            data[:, byte::5] |= ((data[:, 4::5] >> ((4 - byte) * 2)) & 3)
        self.array = np.zeros(
            (data.shape[0], data.shape[1] * 4 // 5), dtype=np.uint16)
        for i in range(4):
            self.array[:, i::4] = data[:, i::5]
        if self.output_dims == 3:
            self.array = self._to_3d(self.array)

    def demosaic(self):
        """
        Perform a rudimentary `de-mosaic`_ of ``self.array``, returning the
        result as a new array. The result of the demosaic is *always* three
        dimensional, with the last dimension being the color planes (see
        *output_dims* parameter on the constructor).

        .. _de-mosaic: https://en.wikipedia.org/wiki/Demosaicing
        """
        if self._demo is None:
            # Construct 3D representation of Bayer data (if necessary)
            if self.output_dims == 2:
                array_3d = self._to_3d(self.array)
            else:
                array_3d = self.array
            # Construct representation of the bayer pattern
            bayer = np.zeros(array_3d.shape, dtype=np.uint8)
            (
                (ry, rx), (gy, gx), (Gy, Gx), (by, bx)
                ) = PiBayerArray.BAYER_OFFSETS[self._header.bayer_order]
            bayer[ry::2, rx::2, 0] = 1 # Red
            bayer[gy::2, gx::2, 1] = 1 # Green
            bayer[Gy::2, Gx::2, 1] = 1 # Green
            bayer[by::2, bx::2, 2] = 1 # Blue
            # Allocate output array with same shape as data and set up some
            # constants to represent the weighted average window
            window = (3, 3)
            borders = (window[0] - 1, window[1] - 1)
            border = (borders[0] // 2, borders[1] // 2)
            # Pad out the data and the bayer pattern (np.pad is faster but
            # unavailable on the version of numpy shipped with Raspbian at the
            # time of writing)
            rgb = np.zeros((
                array_3d.shape[0] + borders[0],
                array_3d.shape[1] + borders[1],
                array_3d.shape[2]), dtype=array_3d.dtype)
            rgb[
                border[0]:rgb.shape[0] - border[0],
                border[1]:rgb.shape[1] - border[1],
                :] = array_3d
            bayer_pad = np.zeros((
                array_3d.shape[0] + borders[0],
                array_3d.shape[1] + borders[1],
                array_3d.shape[2]), dtype=bayer.dtype)
            bayer_pad[
                border[0]:bayer_pad.shape[0] - border[0],
                border[1]:bayer_pad.shape[1] - border[1],
                :] = bayer
            bayer = bayer_pad
            # For each plane in the RGB data, construct a view over the plane
            # of 3x3 matrices. Then do the same for the bayer array and use
            # Einstein summation to get the weighted average
            self._demo = np.empty(array_3d.shape, dtype=array_3d.dtype)
            for plane in range(3):
                p = rgb[..., plane]
                b = bayer[..., plane]
                pview = as_strided(p, shape=(
                    p.shape[0] - borders[0],
                    p.shape[1] - borders[1]) + window, strides=p.strides * 2)
                bview = as_strided(b, shape=(
                    b.shape[0] - borders[0],
                    b.shape[1] - borders[1]) + window, strides=b.strides * 2)
                psum = np.einsum('ijkl->ij', pview)
                bsum = np.einsum('ijkl->ij', bview)
                self._demo[..., plane] = psum // bsum
        return self._demo


class PiMotionArray(PiArrayOutput):
    """
    Produces a 3-dimensional array of motion vectors from the H.264 encoder.

    This custom output class is intended to be used with the *motion_output*
    parameter of the :meth:`~picamera.PiCamera.start_recording` method.  Once
    recording has finished, the class generates a 3-dimensional numpy array
    organized as (frames, rows, columns) where ``rows`` and ``columns`` are the
    number of rows and columns of `macro-blocks`_ (16x16 pixel blocks) in the
    original frames. There is always one extra column of macro-blocks present
    in motion vector data.

    The data-type of the :attr:`~PiArrayOutput.array` is an (x, y, sad)
    structure where ``x`` and ``y`` are signed 1-byte values, and ``sad`` is an
    unsigned 2-byte value representing the `sum of absolute differences`_ of
    the block. For example::

        import picamera
        import picamera.array

        with picamera.PiCamera() as camera:
            with picamera.array.PiMotionArray(camera) as output:
                camera.resolution = (640, 480)
                camera.start_recording(
                      '/dev/null', format='h264', motion_output=output)
                camera.wait_recording(30)
                camera.stop_recording()
                print('Captured %d frames' % output.array.shape[0])
                print('Frames are %dx%d blocks big' % (
                    output.array.shape[2], output.array.shape[1]))

    If you are using the GPU resizer with your recording, use the optional
    *size* parameter to specify the resizer's output resolution when
    constructing the array::

        import picamera
        import picamera.array

        with picamera.PiCamera() as camera:
            camera.resolution = (640, 480)
            with picamera.array.PiMotionArray(camera, size=(320, 240)) as output:
                camera.start_recording(
                    '/dev/null', format='h264', motion_output=output,
                    resize=(320, 240))
                camera.wait_recording(30)
                camera.stop_recording()
                print('Captured %d frames' % output.array.shape[0])
                print('Frames are %dx%d blocks big' % (
                    output.array.shape[2], output.array.shape[1]))

    .. note::

        This class is not suitable for real-time analysis of motion vector
        data. See the :class:`PiMotionAnalysis` class instead.

    .. _macro-blocks: https://en.wikipedia.org/wiki/Macroblock
    .. _sum of absolute differences: https://en.wikipedia.org/wiki/Sum_of_absolute_differences
    """

    def flush(self):
        super(PiMotionArray, self).flush()
        width, height = self.size or self.camera.resolution
        cols = ((width + 15) // 16) + 1
        rows = (height + 15) // 16
        b = self.getvalue()
        frames = len(b) // (cols * rows * motion_dtype.itemsize)
        self.array = np.frombuffer(b, dtype=motion_dtype).reshape((frames, rows, cols))


class PiAnalysisOutput(io.IOBase):
    """
    Base class for analysis outputs.

    This class extends :class:`io.IOBase` with a stub :meth:`analyze` method
    which will be called for each frame output. In this base implementation the
    method simply raises :exc:`NotImplementedError`.
    """

    def __init__(self, camera, size=None):
        super(PiAnalysisOutput, self).__init__()
        self.camera = camera
        self.size = size

    def writable(self):
        return True

    def write(self, b):
        return len(b)

    def analyze(self, array):
        """
        Stub method for users to override.
        """
        try:
            self.analyse(array)
            warnings.warn(
                PiCameraDeprecated(
                    'The analyse method is deprecated; use analyze (US '
                    'English spelling) instead'))
        except NotImplementedError:
            raise

    def analyse(self, array):
        """
        Deprecated alias of :meth:`analyze`.
        """
        raise NotImplementedError


class PiRGBAnalysis(PiAnalysisOutput):
    """
    Provides a basis for per-frame RGB analysis classes.

    This custom output class is intended to be used with the
    :meth:`~picamera.PiCamera.start_recording` method when it is called with
    *format* set to ``'rgb'`` or ``'bgr'``. While recording is in progress, the
    :meth:`~PiAnalysisOutput.write` method converts incoming frame data into a
    numpy array and calls the stub :meth:`~PiAnalysisOutput.analyze` method
    with the resulting array (this deliberately raises
    :exc:`NotImplementedError` in this class; you must override it in your
    descendent class).

    .. note::

        If your overridden :meth:`~PiAnalysisOutput.analyze` method runs slower
        than the required framerate (e.g. 33.333ms when framerate is 30fps)
        then the camera's effective framerate will be reduced. Furthermore,
        this doesn't take into account the overhead of picamera itself so in
        practice your method needs to be a bit faster still.

    The array passed to :meth:`~PiAnalysisOutput.analyze` is organized as
    (rows, columns, channel) where the channels 0, 1, and 2 are R, G, and B
    respectively (or B, G, R if *format* is ``'bgr'``).
    """

    def write(self, b):
        result = super(PiRGBAnalysis, self).write(b)
        self.analyze(bytes_to_rgb(b, self.size or self.camera.resolution))
        return result


class PiYUVAnalysis(PiAnalysisOutput):
    """
    Provides a basis for per-frame YUV analysis classes.

    This custom output class is intended to be used with the
    :meth:`~picamera.PiCamera.start_recording` method when it is called with
    *format* set to ``'yuv'``. While recording is in progress, the
    :meth:`~PiAnalysisOutput.write` method converts incoming frame data into a
    numpy array and calls the stub :meth:`~PiAnalysisOutput.analyze` method
    with the resulting array (this deliberately raises
    :exc:`NotImplementedError` in this class; you must override it in your
    descendent class).

    .. note::

        If your overridden :meth:`~PiAnalysisOutput.analyze` method runs slower
        than the required framerate (e.g. 33.333ms when framerate is 30fps)
        then the camera's effective framerate will be reduced. Furthermore,
        this doesn't take into account the overhead of picamera itself so in
        practice your method needs to be a bit faster still.

    The array passed to :meth:`~PiAnalysisOutput.analyze` is organized as
    (rows, columns, channel) where the channel 0 is Y (luminance), while 1 and
    2 are U and V (chrominance) respectively. The chrominance values normally
    have quarter resolution of the luminance values but this class makes all
    channels equal resolution for ease of use.
    """

    def write(self, b):
        result = super(PiYUVAnalysis, self).write(b)
        self.analyze(bytes_to_yuv(b, self.size or self.camera.resolution))
        return result


class PiMotionAnalysis(PiAnalysisOutput):
    """
    Provides a basis for real-time motion analysis classes.

    This custom output class is intended to be used with the *motion_output*
    parameter of the :meth:`~picamera.PiCamera.start_recording` method.  While
    recording is in progress, the write method converts incoming motion data
    into numpy arrays and calls the stub :meth:`~PiAnalysisOutput.analyze`
    method with the resulting array (which deliberately raises
    :exc:`NotImplementedError` in this class).

    .. note::

        If your overridden :meth:`~PiAnalysisOutput.analyze` method runs slower
        than the required framerate (e.g. 33.333ms when framerate is 30fps)
        then the camera's effective framerate will be reduced. Furthermore,
        this doesn't take into account the overhead of picamera itself so in
        practice your method needs to be a bit faster still.

    The array passed to :meth:`~PiAnalysisOutput.analyze` is organized as
    (rows, columns) where ``rows`` and ``columns`` are the number of rows and
    columns of `macro-blocks`_ (16x16 pixel blocks) in the original frames.
    There is always one extra column of macro-blocks present in motion vector
    data.

    The data-type of the array is an (x, y, sad) structure where ``x`` and
    ``y`` are signed 1-byte values, and ``sad`` is an unsigned 2-byte value
    representing the `sum of absolute differences`_ of the block.

    An example of a crude motion detector is given below::

        import numpy as np
        import picamera
        import picamera.array

        class DetectMotion(picamera.array.PiMotionAnalysis):
            def analyze(self, a):
                a = np.sqrt(
                    np.square(a['x'].astype(np.float)) +
                    np.square(a['y'].astype(np.float))
                    ).clip(0, 255).astype(np.uint8)
                # If there're more than 10 vectors with a magnitude greater
                # than 60, then say we've detected motion
                if (a > 60).sum() > 10:
                    print('Motion detected!')

        with picamera.PiCamera() as camera:
            with DetectMotion(camera) as output:
                camera.resolution = (640, 480)
                camera.start_recording(
                      '/dev/null', format='h264', motion_output=output)
                camera.wait_recording(30)
                camera.stop_recording()

    You can use the optional *size* parameter to specify the output resolution
    of the GPU resizer, if you are using the *resize* parameter of
    :meth:`~picamera.PiCamera.start_recording`.
    """

    def __init__(self, camera, size=None):
        super(PiMotionAnalysis, self).__init__(camera, size)
        self.cols = None
        self.rows = None

    def write(self, b):
        result = super(PiMotionAnalysis, self).write(b)
        if self.cols is None:
            width, height = self.size or self.camera.resolution
            self.cols = ((width + 15) // 16) + 1
            self.rows = (height + 15) // 16
        self.analyze(
                np.frombuffer(b, dtype=motion_dtype).\
                reshape((self.rows, self.cols)))
        return result


class MMALArrayBuffer(mo.MMALBuffer):
    __slots__ = ('_shape',)

    def __init__(self, port, buf):
        super(MMALArrayBuffer, self).__init__(buf)
        width = port._format[0].es[0].video.width
        height = port._format[0].es[0].video.height
        bpp = self.size // (width * height)
        self.offset = 0
        self.length = width * height * bpp
        self._shape = (height, width, bpp)

    def __enter__(self):
        mmal_check(
            mmal.mmal_buffer_header_mem_lock(self._buf),
            prefix='unable to lock buffer header memory')
        assert self.offset == 0
        return np.frombuffer(
            ct.cast(
                self._buf[0].data,
                ct.POINTER(ct.c_uint8 * self._buf[0].alloc_size)).contents,
            dtype=np.uint8, count=self.length).reshape(self._shape)

    def __exit__(self, *exc):
        mmal.mmal_buffer_header_mem_unlock(self._buf)
        return False


class PiArrayTransform(mo.MMALPythonComponent):
    """
    A derivative of :class:`~picamera.mmalobj.MMALPythonComponent` which eases
    the construction of custom MMAL transforms by representing buffer data as
    numpy arrays. The *formats* parameter specifies the accepted input
    formats as a sequence of strings (default: 'rgb', 'bgr', 'rgba', 'bgra').

    Override the :meth:`transform` method to modify buffers sent to the
    component, then place it in your MMAL pipeline as you would a normal
    encoder.
    """
    __slots__ = ()

    def __init__(self, formats=('rgb', 'bgr', 'rgba', 'bgra')):
        super(PiArrayTransform, self).__init__()
        if isinstance(formats, bytes):
            formats = formats.decode('ascii')
        if isinstance(formats, str):
            formats = (formats,)
        try:
            formats = {
                {
                    'rgb': mmal.MMAL_ENCODING_RGB24,
                    'bgr': mmal.MMAL_ENCODING_BGR24,
                    'rgba': mmal.MMAL_ENCODING_RGBA,
                    'bgra': mmal.MMAL_ENCODING_BGRA,
                    }[fmt]
                for fmt in formats
                }
        except KeyError as e:
            raise PiCameraValueError(
                'PiArrayTransform cannot handle format %s' % str(e))
        self.inputs[0].supported_formats = formats
        self.outputs[0].supported_formats = formats

    def _callback(self, port, source_buf):
        try:
            target_buf = self.outputs[0].get_buffer(False)
        except PiCameraPortDisabled:
            return False
        if target_buf:
            target_buf.copy_meta(source_buf)
            result = self.transform(
                MMALArrayBuffer(port, source_buf._buf),
                MMALArrayBuffer(self.outputs[0], target_buf._buf))
            try:
                self.outputs[0].send_buffer(target_buf)
            except PiCameraPortDisabled:
                return False
        return False

    def transform(self, source, target):
        """
        This method will be called for every frame passing through the
        transform.  The *source* and *target* parameters represent buffers from
        the input and output ports of the transform respectively. They will be
        derivatives of :class:`~picamera.mmalobj.MMALBuffer` which return a
        3-dimensional numpy array when used as context managers. For example::

            def transform(self, source, target):
                with source as source_array, target as target_array:
                    # Copy the source array data to the target
                    target_array[...] = source_array
                    # Draw a box around the edges
                    target_array[0, :, :] = 0xff
                    target_array[-1, :, :] = 0xff
                    target_array[:, 0, :] = 0xff
                    target_array[:, -1, :] = 0xff
                    return False

        The target buffer's meta-data starts out as a copy of the source
        buffer's meta-data, but the target buffer's data starts out
        uninitialized.
        """
        return False

